/*
This file is part of [aconnect] library. 

Author: Artem Kustikov (kustikoff[at]tut.by)


This code is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any
damages arising from the use of this code.

Permission is granted to anyone to use this code for any
purpose, including commercial applications, and to alter it and
redistribute it freely, subject to the following restrictions:

1. The origin of this code must not be misrepresented; you must
not claim that you wrote the original code. If you use this
code in a product, an acknowledgment in the product documentation
would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and
must not be misrepresented as being the original code.

3. This notice may not be removed or altered from any source
distribution.
*/

#include "config.hpp"

#include <boost/algorithm/string.hpp>
#include <boost/filesystem.hpp>

#include <iostream>

#include "util.hpp"
#include "util.string.hpp"
#include "util.time.hpp"
#include "logger.hpp"

namespace aconnect
{
	bool Logger::valid()	{	
		return true; 
	}


	void Logger::processMessage (Log::LogLevel level, string_constptr msg)
	{
		if (level > _level)
			return;

		if(!valid()) {
			if (_useStderrTillInit)
				std::cerr << msg;
			return;
		}
		
		const int bufferSize = 60;
		aconnect::char_type buff[bufferSize] = {0};

		// record example
		// [27-12-2008 01:04:07]   3076 Info: Server started

		struct tm tmTime = util::getDateTime();
		
		string res (msg);
		res.reserve (res.size() + bufferSize);

		string_constptr prefix = NULL;

		switch (level) {
			case Log::Debug:
				prefix = Log::DebugMsg; break;
			case Log::Info:
				prefix = Log::InfoMsg; break;
			case Log::Warning:
				prefix = Log::WarningMsg; break;
			case Log::Error:
				prefix = Log::ErrorMsg; break;
			case Log::Critical:
				prefix = Log::CriticalMsg; break;

			default:
				prefix = Log::UnknownMsg;
		}
		

		int formatted = snprintf (buff, bufferSize, "%02d-%02d-%04d %02d:%02d:%02d (%s) %s: ", 
			tmTime.tm_mday,
			(tmTime.tm_mon + 1),
			(tmTime.tm_year + 1900),
			tmTime.tm_hour,
			tmTime.tm_min,
			tmTime.tm_sec,
			boost::lexical_cast<string> (util::getCurrentThreadId()).c_str(),
			prefix
		);

		assert (formatted > 0);
		
		res.insert(0, buff, formatted); 
		writeMessage ( res );
	}

	//////////////////////////////////////////////////////////////////////////
	//
	//		ConsoleLogger
	//
	void ConsoleLogger::writeMessage (string_constref msg)
	{
		boost::mutex::scoped_lock lock (_writeMessageMutex);
		std::cout << msg << std::endl;
	}

	//////////////////////////////////////////////////////////////////////////
	//
	//		FileLogger
	//

	bool FileLogger::valid()
	{	
		return _output.is_open() && !_output.fail(); 
	}

	void FileLogger::init (Log::LogLevel level, string_constptr filePathTemplate, 
	        size_t maxFileSize, bool createFile)  throw (std::runtime_error) 
	{
		_level = level;
		_maxFileSize = maxFileSize;

		if (util::isNullOrEmpty (filePathTemplate))
			throw std::runtime_error ("Log file name template is null or empty");

		if (_useStderrTillInit)
			_useStderrTillInit = false;

		_filePathTemplate = filePathTemplate;
	
		if (createFile)
    		createLogFile ();
	}

	FileLogger::~FileLogger() {
		destroy();
	}

	void FileLogger::closeWriter () {
		if (valid()) 
			_output.flush();
		_output.rdbuf()->close();
	}

	void FileLogger::destroy() 
	{
		boost::mutex::scoped_lock lock (_writeMessageMutex);
		closeWriter ();
		_output.close();
	}

	string FileLogger::generateTimeStamp() {
		
		struct tm tmTime = util::getDateTime();
		
		const int bufferSize = 24;
		aconnect::char_type buff[bufferSize] = {0};

		int formatted = snprintf (buff, bufferSize, "%04d_%02d_%02d_%02d_%02d_%02d", 
			(tmTime.tm_year + 1900),
			(tmTime.tm_mon + 1),
			tmTime.tm_mday,
			(tmTime.tm_hour),
			(tmTime.tm_min),
			(tmTime.tm_sec)
		);

		assert (formatted > 0);

		return (formatted > 0 ? buff : "0000000"); 
	}


	void FileLogger::createLogFile()  throw (std::runtime_error)
	{
		namespace fs = boost::filesystem;
		
		string ts = generateTimeStamp();
		string fileNameInit;
		if (_filePathTemplate.find(Log::TimeStampMark) != string::npos ) 
			fileNameInit = boost::algorithm::replace_all_copy(_filePathTemplate, Log::TimeStampMark, ts);
		else
			fileNameInit = _filePathTemplate + ts;
		
		fs::path fileName (fileNameInit);
		const string ext = fs::extension (fileName);

		const int bufferSize = 64;
		aconnect::char_type buff[bufferSize] = {0};

		string formatStr = ".%06d" + ext.substr(0, bufferSize-8);
		
		int ndx = 0;
		while ( fs::exists (fileName) ) {
			
			int formatted = snprintf (buff, bufferSize, formatStr.c_str(), ndx);
			assert (formatted > 0);

			fileName = fs::change_extension (fs::path (fileNameInit), buff) ;
			++ndx;
		}

		closeWriter ();

		_currentFilePath = fileName.file_string();

		_output.open ( _currentFilePath.c_str(), std::ios::out | std::ios::binary );

		if (_output.fail())
			throw application_error ("Cannot create \"%s\" log file", fileName.file_string().c_str() );
		
	    // _output << "---------------------------------------" << std::endl;
	}

	void FileLogger::writeMessage (string_constref msg)
	{
		boost::mutex::scoped_lock lock (_writeMessageMutex);
		if(!valid())
			return;
		
		writeMessageToFile (msg);
	}

	void FileLogger::writeMessageToFile (string_constref msg) throw (std::runtime_error) 
	{
		_outputSize += msg.size();
		_output << msg << Log::EOL;

		if (_output.fail())
			throw std::runtime_error ("Error writing log file");

		if (_outputSize >= _maxFileSize) {
			createLogFile ();
			_outputSize = 0;
		}

	}

	//////////////////////////////////////////////////////////////////////////
	//
	//		BackgroundFileLogger
	//
	BackgroundFileLogger::BackgroundFileLogger ():
        _writerThread (),
		_messagesBufferSize (0),
		_maxBufferSize (1024 * 1024),
		_flushWaitTimeout (1), // sec
		_active (true)
	{
		
	}

	bool BackgroundFileLogger::valid()
	{	
		return FileLogger::valid() && isActive(); 
	}

	void BackgroundFileLogger::init (Log::LogLevel level, string_constptr filePathTemplate,
			size_t maxFileSize, bool createFile) throw (std::runtime_error)
	{
	    FileLogger::init (level, filePathTemplate, maxFileSize, createFile);   
	    
	    _writerThread.reset (new boost::thread (thread_proc_adapter<void (*)(BackgroundFileLogger*), BackgroundFileLogger*>
				(BackgroundFileLogger::run, this) ));

        boost::thread::yield ();
		
	}
			
	void BackgroundFileLogger::run (BackgroundFileLogger* backgroundLogger) 
	{
		backgroundLogger->prepare();

        while (backgroundLogger->doWaitAndFlush());
	}

	void BackgroundFileLogger::prepare () 
	{
		// boost::this_thread::sleep (boost::posix_time::milliseconds(_flushWaitTimeout * 1000));
	}

	bool BackgroundFileLogger::doWaitAndFlush () 
	{
		std::deque<string> messagesToFlush;
		bool notified = false;
		
		// load messages
		try 
		{

			boost::mutex::scoped_lock lock (_flushMutex);
			notified = _flushCondition.timed_wait (lock, util::createTimePeriod(_flushWaitTimeout) );
	
			if (!isActive())
				return false;

			{
				boost::mutex::scoped_lock writeLock (_writeMessageMutex);

				_messages.swap (messagesToFlush);
				_messagesBufferSize = 0;
			}

			boost::this_thread::interruption_point();

			writeQueue (messagesToFlush);
		
		} catch (boost::thread_resource_error) {
			
			// eat exceptions at destroy()
			if (notified)
				throw;
		}
		
		return _active;
	}

	void BackgroundFileLogger::writeMessage(string_constref msg)
	{
		boost::mutex::scoped_lock lock (_writeMessageMutex);
		if(!isActive())
		    return;
		
		_messagesBufferSize += msg.size();
		_messages.push_back(msg);

		if (_messagesBufferSize >= _maxBufferSize)
			_flushCondition.notify_one();
	}
	

	void BackgroundFileLogger::writeQueue (std::deque<string>& messages ) 
	{
		while (messages.size() > 0) 
		{
			writeMessageToFile (messages.front() );
			messages.pop_front();
		}

		_output.flush ();
	}


	void BackgroundFileLogger::destroy()
	{
		boost::mutex::scoped_lock lock (_flushMutex);
		_active = false;
        
		_flushCondition.notify_one(); // free write thread

		// flush last messages
		writeQueue (_messages);

		lock.unlock();
	
		if (_writerThread.get())
			_writerThread->interrupt();

		FileLogger::destroy();
	}
	
	//////////////////////////////////////////////////////////////////////////
	//
	//		ProgressTimer
	//

	ProgressTimer::~ProgressTimer () {
		try 
		{
			const int bufferSize = 64;
			aconnect::char_type buff[bufferSize] = {0};

			int formatted = snprintf (buff, bufferSize, 
				"%s: elapsed time - %f sec", 
				_funcName.c_str(),
				_timer.elapsed()
			);
			if (formatted > 0 ) 
			{
				if (formatted == bufferSize)
					buff [bufferSize - 1] = '\0';
				
				_log.processMessage (_level, buff);
			}

		} catch (...) {
			// eat exception
		}
	}

}
