/*
This file is part of [ahttp] library. 

Author: Artem Kustikov (kustikoff[at]tut.by)


This code is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any
damages arising from the use of this code.

Permission is granted to anyone to use this code for any
purpose, including commercial applications, and to alter it and
redistribute it freely, subject to the following restrictions:

1. The origin of this code must not be misrepresented; you must
not claim that you wrote the original code. If you use this
code in a product, an acknowledgment in the product documentation
would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and
must not be misrepresented as being the original code.

3. This notice may not be removed or altered from any source
distribution.
*/


#include "aconnect/config.hpp"
#include "aconnect/boost_format_safe.hpp"

#include <boost/filesystem.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>

#include <assert.h>

#include "aconnect/util.hpp"

#include "ahttp/http_support.hpp"
#include "ahttp/http_server_settings.hpp"

#if defined(__GNUC__)
#	include <dlfcn.h>
#endif

namespace algo = boost::algorithm;
namespace fs = boost::filesystem;

#include "http_messages.inl"

namespace ahttp
{
	DirectorySettings::DirectorySettings () : 
		number (0),
		browsingEnabled (Tristate::Undefined), 
		showHiddenFiles (Tristate::Undefined), 
		isVirtualDirectory(false),
		maxRequestSize (-1),
		enableParentPathAccess (Tristate::Undefined),
		parentServerSettings (NULL)
	{

	}

	void DirectorySettings::processTemplates()
	{
		// add tabulators and EOLs
		algo::replace_all (headerTemplate, ahttp::SettingsConstants::TabulatorMark, "\t");
		algo::replace_all (parentDirectoryTemplate, ahttp::SettingsConstants::TabulatorMark, "\t");
		algo::replace_all (fileTemplate, ahttp::SettingsConstants::TabulatorMark, "\t");
		algo::replace_all (directoryTemplate, ahttp::SettingsConstants::TabulatorMark, "\t");
		algo::replace_all (virtualDirectoryTemplate, ahttp::SettingsConstants::TabulatorMark, "\t");
		algo::replace_all (footerTemplate, ahttp::SettingsConstants::TabulatorMark, "\t");

		algo::replace_all (headerTemplate, ahttp::SettingsConstants::EndOfLineMark, "\r\n");
		algo::replace_all (parentDirectoryTemplate, ahttp::SettingsConstants::EndOfLineMark, "\r\n");
		algo::replace_all (fileTemplate, ahttp::SettingsConstants::EndOfLineMark, "\r\n");
		algo::replace_all (directoryTemplate, ahttp::SettingsConstants::EndOfLineMark, "\r\n");
		algo::replace_all (virtualDirectoryTemplate, ahttp::SettingsConstants::EndOfLineMark, "\r\n");
		algo::replace_all (footerTemplate, ahttp::SettingsConstants::EndOfLineMark, "\r\n");
	}

	HttpServerSettings::HttpServerSettings() :
		InstanceId (1),
		IsKeepAliveEnabled (settings_defaults::EnableKeepAlive),
		KeepAliveTimeout (settings_defaults::KeepAliveTimeout),
		IsGzipDeflateEnabled (settings_defaults::EnableGzipDeflate),
		CommandSocketTimeout (settings_defaults::CommandSocketTimeout),
		UploadCreationTriesCount (settings_defaults::UploadCreationTriesCount),
		ResponseBufferSize (settings_defaults::ResponseBufferSize),
		MaxChunkSize (settings_defaults::MaxChunkSize),

		_port(-1), 
		_commandPort (-1),
		_logger (NULL),
		_serverVersion (settings_defaults::ServerVersion)
		
	{
		ConnectionSettings.socketReadTimeout = settings_defaults::ServerSocketTimeout;
		ConnectionSettings.socketWriteTimeout = settings_defaults::ServerSocketTimeout;

		initMessages ();
	}

	HttpServerSettings::~HttpServerSettings()
	{
	}

	string HttpServerSettings::getMimeType (string_constref ext) const 
	{
		using namespace aconnect;
		
		str2str_map::const_iterator it = _mimeTypes.find(ext);
		if (it != _mimeTypes.end())
			return it->second;

		return strings::ContentTypeOctetStream;			
	}

	void HttpServerSettings::loadMimeType (string_constref ext, string_constptr typeDesc)
	{
		if (typeDesc)
			_mimeTypes [ext] = typeDesc;
	}

	const DirectorySettings* HttpServerSettings::getDirSettingsByName(string_constref dirName) const	
	{
		directories_map::const_iterator it = _directories.begin();

		for (;it != _directories.end(); ++it) {
			if (aconnect::util::equals(it->second.name, dirName)) 
				return &it->second;
		}
		
		return 0;			
	}

	const DirectorySettings* HttpServerSettings::getDirSettings(int number) const	
	{
		directories_map::const_iterator it = _directories.begin();

		for (;it != _directories.end(); ++it) {
			if (it->second.number == number)
				return &it->second;
		}
		
		return 0;			
	}

	void HttpServerSettings::setupRootDirectory (DirectorySettings &directoryInfo,
		boost::function1<void, DirectorySettings* > locaInfoLoadCallback) throw (settings_load_error)
	{
		fs::path rootPath = fs::path(directoryInfo.realPath, fs::portable_directory_name);

		if ( !fs::exists (rootPath) )
			throw settings_load_error ("Incorrect path in root directory record - path does not exist: %s", 
				rootPath.directory_string().c_str() );
		
		if ( !fs::is_directory (rootPath) )
			throw settings_load_error ("Incorrect path in root directory record - target is not a directory");

		directoryInfo.realPath = rootPath.directory_string();
		directoryInfo.virtualPath = strings::Slash;

		// try to load local directory configuration
		if (locaInfoLoadCallback != 0)
			locaInfoLoadCallback (&directoryInfo);
		
		// setup defaults
		if (directoryInfo.maxRequestSize == (size_t) -1) // is not loaded
			directoryInfo.maxRequestSize = settings_defaults::MaxRequestSize;

		if (directoryInfo.browsingEnabled == Tristate::Undefined) directoryInfo.browsingEnabled = Tristate::False;
		if (directoryInfo.showHiddenFiles == Tristate::Undefined) directoryInfo.showHiddenFiles = Tristate::False;

		// register root
		directoryInfo.number = 0;
		
		// apply default templates
		if (directoryInfo.headerTemplate.empty()) directoryInfo.headerTemplate = settings_defaults::HeaderTemplate;
		if (directoryInfo.parentDirectoryTemplate.empty()) directoryInfo.parentDirectoryTemplate = settings_defaults::ParentDirectoryTemplate;
		if (directoryInfo.directoryTemplate.empty()) directoryInfo.directoryTemplate = settings_defaults::DirectoryTemplate;
		if (directoryInfo.virtualDirectoryTemplate.empty()) directoryInfo.virtualDirectoryTemplate = settings_defaults::VirtualDirectoryTemplate;
		if (directoryInfo.fileTemplate.empty()) directoryInfo.fileTemplate = settings_defaults::FileTemplate;
		if (directoryInfo.footerTemplate.empty()) directoryInfo.footerTemplate = settings_defaults::FooterTemplate;	
		directoryInfo.processTemplates();
	
		loadGlobalModules (directoryInfo);
		
		// copy to internal map
		_directories[directoryInfo.virtualPath] = directoryInfo;

		fillModulesCallbackInfo();
	}

	void HttpServerSettings::loadRootDirectory (std::vector <DirectorySettings>& directoriesList, 
			boost::function1<void, DirectorySettings* > locaInfoLoadCallback) throw (settings_load_error)
	{
		
		std::vector <DirectorySettings>::iterator it = directoriesList.end();
		for (it = directoriesList.begin(); 
			it != directoriesList.end() && !util::equals(it->name, _rootDirName); ++it );

		if (it == directoriesList.end())
			throw settings_load_error ("There is no directory record with name \"%s\"", 
				_rootDirName.c_str());

		if (it->realPath.empty())
			throw settings_load_error ("Empty path in root directory record");

		// work with filesystem
		try
		{
			setupRootDirectory (*it, locaInfoLoadCallback);

			fillDirectoriesMap (directoriesList, it, locaInfoLoadCallback);

			fillModulesCallbackInfo();

		} catch (fs::basic_filesystem_error<fs::path> &err) {
			throw settings_load_error (
				"Directories info loading failed - 'basic_filesystem_error' caught: %s, "
				"path 1: %s, path 2: %s", 
				err.what(), 
				err.path1().string().c_str(),
				err.path2().string().c_str()); 
		} catch (std::exception &ex)  {
			throw settings_load_error ("Directories info loading failed - exception [%s]: %s", 
				typeid(ex).name(), ex.what());;

		} catch (...)  {
			throw settings_load_error ("Directories info loading failed - unknown exception caught");
		}
	}

	
	void HttpServerSettings::fillDirectoriesMap (std::vector<ahttp::DirectorySettings>& directoriesList, 
												 std::vector<ahttp::DirectorySettings>::iterator parent,
												 boost::function1<void, DirectorySettings* > locaInfoLoadCallback)
	{
		using namespace aconnect;
		
		std::vector <DirectorySettings>::iterator childIter;
		for (childIter = directoriesList.begin(); childIter != directoriesList.end(); ++childIter ) 
		{
			if (childIter->parentName == parent->name) 
			{
				if (childIter->virtualPath.empty())
					throw settings_load_error ("Empty <virtual-path> for nested directory: %s", 
					childIter->name.c_str());
				
				string virtualPathInit = childIter->virtualPath;

				childIter->virtualPath = parent->virtualPath + childIter->virtualPath;
				if (childIter->virtualPath.substr(childIter->virtualPath.length() - 1) != strings::Slash)
					childIter->virtualPath += strings::Slash;

				// get real path
				fs::path childPath;
				if (!childIter->realPath.empty()) {
					childPath = fs::path (childIter->realPath, fs::portable_directory_name);
					childIter->isVirtualDirectory = true;
					childIter->relativePath = virtualPathInit;

					if ( !fs::exists (childPath) )
						throw settings_load_error ("Incorrect path in <directory> record - path does not exist,\
												   directory: %s", childIter->name.c_str());
					if ( !fs::is_directory (childPath) )
						throw settings_load_error ("Incorrect path in <directory> record - target is not a directory,\
													directory: %s", childIter->name.c_str());
				
				} else if (!childIter->relativePath.empty()) {
					childPath = fs::complete (fs::path (childIter->relativePath, fs::portable_directory_name),
						fs::path (parent->realPath, fs::native));
				
				} else {
					childIter->isVirtualDirectory = true;
					childIter->relativePath = virtualPathInit;

					childPath = fs::complete (fs::path (virtualPathInit, fs::portable_directory_name),
						fs::path (parent->realPath, fs::native));
				}

				if (fs::exists (childPath) && fs::is_directory (childPath))
				{
					childIter->realPath = childPath.directory_string();
					// try to load local directory configuration
					if (locaInfoLoadCallback != 0)
						locaInfoLoadCallback ( &(*childIter) );
				}
				
				// copy properties
				if (childIter->browsingEnabled == Tristate::Undefined) childIter->browsingEnabled = parent->browsingEnabled;
				if (childIter->showHiddenFiles == Tristate::Undefined) childIter->showHiddenFiles = parent->showHiddenFiles;

				if (childIter->charset.empty()) childIter->charset = parent->charset;
				if (childIter->maxRequestSize == (size_t) -1) childIter->maxRequestSize = parent->maxRequestSize;
				if (childIter->enableParentPathAccess == Tristate::Undefined) childIter->enableParentPathAccess = parent->enableParentPathAccess;

				if (childIter->headerTemplate.empty())	childIter->headerTemplate = parent->headerTemplate;
				if (childIter->parentDirectoryTemplate.empty())	childIter->parentDirectoryTemplate = parent->parentDirectoryTemplate;
				if (childIter->fileTemplate.empty())	childIter->fileTemplate = parent->fileTemplate;
				if (childIter->directoryTemplate.empty())	childIter->directoryTemplate = parent->directoryTemplate;
				if (childIter->virtualDirectoryTemplate.empty())	childIter->virtualDirectoryTemplate = parent->virtualDirectoryTemplate;
				if (childIter->footerTemplate.empty())	childIter->footerTemplate = parent->footerTemplate;

				// fill default documents list
				default_documents_vector defDocsList = parent->defaultDocuments;

				default_documents_vector::iterator iter = childIter->defaultDocuments.begin(), removeIter;
				while (iter != childIter->defaultDocuments.end()) 
				{
					if (iter->first && std::find (defDocsList.begin(), defDocsList.end(), *iter) == defDocsList.end()) {
						defDocsList.push_back(*iter);

					} else if (!iter->first) {
						removeIter = std::find (defDocsList.begin(), defDocsList.end(), std::make_pair(true, iter->second));
						if (removeIter == defDocsList.end())
							throw settings_load_error ("Cannot remove default document record \"%s\", in directory: %s - "
							"it is not declared in parent directory record.", 
							iter->second.c_str(), 
							childIter->name.c_str());
						else
							defDocsList.erase(removeIter);
					}
					iter++;
				}
				childIter->defaultDocuments = defDocsList;

				// copy plugins registration
				copyHandlersRegistration (*parent, *childIter);

				copyModulesRegistration (*parent, *childIter);

				// store updated directory info
				childIter->number = _directories.size();
				_directories[childIter->virtualPath] = *childIter;

				fillDirectoriesMap (directoriesList, childIter, locaInfoLoadCallback);
			}

		}

	}


	
	void HttpServerSettings::loadGlobalModules (DirectorySettings& root)
	{
		// load global modules
		global_plugins_map::const_iterator moduleIt = RegisteredModules.begin();
		
		for (; moduleIt != RegisteredModules.end(); ++moduleIt) 
		{
			if (moduleIt->second.global) 
			{
				const PluginInfo& info = moduleIt->second;
				
				PluginRegistrationInfo regInfo;

				regInfo.registrationType = PluginRegistration::Add;
				regInfo.pluginName = moduleIt->first;
				regInfo.pluginIndex = info.pluginIndex;

				std::copy (&info.moduleCallbacks[0], 
					&info.moduleCallbacks[0] + ModuleCallbackTypesCount,
					&regInfo.moduleCallbacks[0]);
		
				root.modules.push_back ( regInfo );
			}
		}
	

	}

	void HttpServerSettings::addModuleCallbacksToMap (void** moduleCallbacks, int pluginIndex, callback_map& callbackMap )
	{
		for (int ndx=0; ndx<ModuleCallbackTypesCount; ++ndx)
		{
			if (moduleCallbacks[ndx])
			{
				callback_map::iterator it = callbackMap.find ((ModuleCallbackType) ndx);
				if (it == callbackMap.end()) {
					std::list< std::pair <void*, int> > lst;
					lst.push_back ( std::make_pair (moduleCallbacks[ndx], pluginIndex));

					callbackMap.insert ( std::make_pair ((ModuleCallbackType) ndx, lst)); 
				
				} else {
					it->second.push_back ( std::make_pair (moduleCallbacks[ndx], pluginIndex) );
				}
			}
		
		}
	}


	void HttpServerSettings::fillModulesCallbackInfo ()
	{
		_modulesCallbacks.clear();
	
		directories_map::const_iterator dirIter = _directories.begin();
	
		for (; dirIter != _directories.end(); ++dirIter) 
		{
			callback_map dirCallbacks;

			directory_plugins_list::const_iterator it = dirIter->second.modules.begin();
			for (; it != dirIter->second.modules.end(); ++it) 
			{
				addModuleCallbacksToMap ( (void**) it->moduleCallbacks, it->pluginIndex, dirCallbacks);
			}
			
			_modulesCallbacks[dirIter->second.number] = dirCallbacks;
		}
	}


	void HttpServerSettings::copyHandlersRegistration (const DirectorySettings& parent, DirectorySettings& child)
	{
		directory_plugins_list registeredHandlers = parent.handlers;
		directory_plugins_list::const_iterator handlerIter;

		for (handlerIter = child.handlers.begin(); 
				handlerIter != child.handlers.end(); ++handlerIter) 
		{
			if (handlerIter->registrationType == PluginRegistration::Clear) {
				registeredHandlers.clear();
			
			} else if (handlerIter->registrationType == PluginRegistration::Remove) {
				
				// if no 'ext' attribute defined - remove registration at all
				if (handlerIter->extensions.empty()) {
					registeredHandlers.erase (std::remove_if (registeredHandlers.begin(), 
								registeredHandlers.end(), PluginInfoByNameFinder(handlerIter->pluginName) ), 
							registeredHandlers.end());

				} else {
					directory_plugins_list::iterator it = std::find_if (registeredHandlers.begin(), 
						registeredHandlers.end(), PluginInfoByNameFinder(handlerIter->pluginName) );

					if (it != registeredHandlers.end()) 
					{
					    std::set<string>::const_iterator itExt;
					    
					    for (itExt = handlerIter->extensions.begin(); itExt != handlerIter->extensions.end(); ++itExt)  {
					        if ( it->extensions.find (*itExt) != it->extensions.end())
					            it->extensions.erase (*itExt);
					    }
				
						it->excludedExtensions.insert (handlerIter->extensions.begin(), handlerIter->extensions.end());
					}
				}
			
			} else if (handlerIter->registrationType == PluginRegistration::Add) {
				
				directory_plugins_list::iterator it = std::find_if (registeredHandlers.begin(), 
						registeredHandlers.end(), PluginInfoByNameFinder(handlerIter->pluginName) );

				if (it == registeredHandlers.end()) {
					registeredHandlers.push_back(*handlerIter);
				} else {
					it->extensions.insert (handlerIter->extensions.begin(), handlerIter->extensions.end());
				}
			
			} else if (handlerIter->registrationType == PluginRegistration::Register) {
				// not safe, but will give some freedom of action for user	
				registeredHandlers.push_back(*handlerIter);
			} 
		}

		// apply handlers info intersection
		child.handlers.swap(registeredHandlers);
	}

	void HttpServerSettings::copyModulesRegistration (const DirectorySettings& parent, DirectorySettings& child)
	{
		directory_plugins_list registeredModules = parent.modules;
		
		directory_plugins_list::const_iterator iter;

		for (iter = child.modules.begin(); 
				iter != child.modules.end(); ++iter) 
		{
			if (iter->registrationType == PluginRegistration::Clear) {
				registeredModules.clear();
			
			} else if (iter->registrationType == PluginRegistration::Remove) {
				
				registeredModules.erase (std::remove_if (registeredModules.begin(), 
					registeredModules.end(), PluginInfoByNameFinder(iter->pluginName) ), 
							registeredModules.end());
			
			} else if (iter->registrationType == PluginRegistration::Add
				|| iter->registrationType == PluginRegistration::Register) {
		
				registeredModules.push_back(*iter);
			} 
		}

		// apply modules info intersection
		child.modules.swap(registeredModules);
	}

	
	void HttpServerSettings::loadPluginProperties (string_constref pluginName, 
		PluginInfo& info, 
		PluginType pluginType) throw (settings_load_error)
	{
		global_plugins_map& pluginsMap = (pluginType == PluginHandler ? RegisteredHandlers : RegisteredModules);
		if (pluginsMap.find (pluginName) != pluginsMap.end() )
			throw settings_load_error ("Plugin \"%s\" has been already loaded", pluginName.c_str());

#if defined (WIN32)		
		info.dll = ::LoadLibraryA (info.pathToLoad.c_str());

		if (NULL == info.dll)
			throw settings_load_error ("Plugin loading failed, library: %s, error: %s", 
				info.pathToLoad.c_str(),
				aconnect::util::formatWindowsError(::GetLastError()).c_str());

		if (pluginType == PluginHandler)
		{
			info.processRequestFunc = ::GetProcAddress (info.dll, SettingsConstants::ProcessRequestFunName);
			if (!info.processRequestFunc) 
				throw settings_load_error ("Request processing function loading failed, "
					"library: %s, error code: %d", info.pathToLoad.c_str(), ::GetLastError());
		}
		
		info.initFunc = ::GetProcAddress (info.dll, SettingsConstants::InitFunName);
		if (!info.initFunc) 
			throw settings_load_error ("Plugin initialization function loading failed, "
				"library: %s, error code: %d", info.pathToLoad.c_str(), ::GetLastError());

		info.destroyFunc = ::GetProcAddress (info.dll, SettingsConstants::DestroyFunName);

#else
		info.dll = dlopen (info.pathToLoad.c_str(), RTLD_NOW | RTLD_LOCAL | RTLD_DEEPBIND );
		if (NULL == info.dll) {
			string_constptr errorMsg = dlerror (); 
			throw settings_load_error ("Plugin loading failed, library: %s, error: %s", 
				info.pathToLoad.c_str(), errorMsg ? errorMsg : "Unknown error");
		}

		if (pluginType == PluginHandler)
		{
			info.processRequestFunc = dlsym (info.dll, SettingsConstants::ProcessRequestFunName);
			if (!info.processRequestFunc) 
				throw settings_load_error ("Request processing function loading failed, "
					"library: %s, error: %s", info.pathToLoad.c_str(), dlerror());
		}

		info.initFunc = dlsym (info.dll, SettingsConstants::InitFunName);
		if (!info.initFunc) 
			throw settings_load_error ("Plugin initialization function loading failed, "
				"library: %s, error: %s", info.pathToLoad.c_str(), dlerror());

		info.destroyFunc = dlsym (info.dll, SettingsConstants::DestroyFunName);
#endif
		if (pluginType == PluginModule)
		{
			info.moduleCallbacks[ModuleCallbackOnRequestBegin] = aconnect::util::getDllSymbol (info.dll, SettingsConstants::OnRequestBeginFunName);
			info.moduleCallbacks[ModuleCallbackOnRequestResolve] = aconnect::util::getDllSymbol (info.dll, SettingsConstants::OnRequestResolveFunName);
			info.moduleCallbacks[ModuleCallbackOnRequestMapHandler] = aconnect::util::getDllSymbol (info.dll, SettingsConstants::OnRequestMapHandlerName);
			
			info.moduleCallbacks[ModuleCallbackOnResponsePreSendHeaders] = aconnect::util::getDllSymbol (info.dll, SettingsConstants::OnResponsePreSendHeadersFunName);
			info.moduleCallbacks[ModuleCallbackOnResponsePreSendContent] = aconnect::util::getDllSymbol (info.dll, SettingsConstants::OnResponsePreSendContentFunName);
			info.moduleCallbacks[ModuleCallbackOnResponseEnd] = aconnect::util::getDllSymbol (info.dll, SettingsConstants::OnResponseEndFunName);
		}

		if (pluginType == PluginHandler)
			RegisteredHandlers[pluginName] = info;
		else if (pluginType == PluginModule)
			RegisteredModules[pluginName] = info;
	}


	void HttpServerSettings::initPlugins (PluginType pluginType, HttpServer* server)
	{
		assert (_logger);
		assert (server);

		global_plugins_map::const_iterator iter;
		global_plugins_map& plugins = (pluginType == PluginHandler ? RegisteredHandlers : RegisteredModules);

		for (iter = plugins.begin(); iter != plugins.end(); ++iter)
		{
			bool inited = reinterpret_cast<init_plugin_function> (iter->second.initFunc) 
				(iter->second.params, iter->second.pluginIndex, server);
			
			if (!inited)
				throw settings_load_error ("Plugin \"%s\" initialization failed",
					iter->first.c_str() );
		}
	}

	void HttpServerSettings::destroyPlugins (PluginType pluginType, bool unload)
	{
		global_plugins_map::iterator iter;
		global_plugins_map& plugins = (pluginType == PluginHandler ? RegisteredHandlers : RegisteredModules);

		for (iter = plugins.begin(); 
			iter != plugins.end(); ++iter)
		{
			assert (iter->second.dll && "DLL is already unloaded");
			if (NULL == iter->second.destroyFunc)
				continue;
			
			reinterpret_cast<destroy_plugin_function> (iter->second.destroyFunc) ();

			if (unload) {
				aconnect::util::unloadLibrary (iter->second.dll);
				iter->second.dll = NULL;
			}
		}
	}

	//==================================================================//
	//	Environment preparation											//
	//==================================================================//
	
	void HttpServerSettings::applyLocale (string_constref localeName)
	{
		char* res = setlocale (LC_CTYPE, localeName.c_str());
		if (NULL == res)
			throw settings_load_error ("Unsupported locale defined: %s", localeName.c_str());
	}

	void HttpServerSettings::setupGlobalUploadsDir ()
	{
		if (!GlobalUploadsDirectory.empty()) 
		{
			updateAppLocationInPath (GlobalUploadsDirectory);

			fs::path uploadsDir (GlobalUploadsDirectory, fs::native); 

			if (!fs::exists (uploadsDir)) {
				try {
					fs::create_directories(uploadsDir);

				} catch (std::runtime_error& err) {
					throw settings_load_error("Upload directory creation failed, exception: %s, message: %s", 
						typeid(err).name(), err.what());
				}
			}
		}
		

	}

	//==================================================================//
	//		Helpers														//
	//==================================================================//

	std::set<string> HttpServerSettings::parseExtensions(string_constref ext) 
	{

		std::set<string> res;
		aconnect::SimpleTokenizer tokens(ext, ";");
		
		for (aconnect::SimpleTokenizer::iterator tok_iter = tokens.begin(); 
				tok_iter != tokens.end(); 
				++tok_iter)
		{
			string token = tok_iter.current_token();
			algo::trim (token);

			if (token.empty())
				continue;

			res.insert (token);
		}

			
		return res;
	}

	void HttpServerSettings::updateAppLocationInPath (aconnect::string &pathStr) const 
	{
		if ( pathStr.find(SettingsConstants::AppPathMark) != aconnect::string::npos)
			algo::replace_first (pathStr, SettingsConstants::AppPathMark, AppLocation);
	}

	//==================================================================//
	//		Defaults													//
	//==================================================================//
	
	void HttpServerSettings::setupDefaultMimeTypes ()
	{
		loadMimeType (".", "application/octet-stream");
		loadMimeType (".ai", "application/postscript");
		loadMimeType (".art", "image/x-jg");
		loadMimeType (".asf", "video/x-ms-asf");
		loadMimeType (".asm", "text/x-asm");
		loadMimeType (".asp", "text/asp");
		loadMimeType (".aspx", "text/aspx");
		loadMimeType (".ascx", "text/aspx");
		loadMimeType (".asx", "video/x-ms-asf");
		loadMimeType (".au", "audio/x-au");
		loadMimeType (".avi", "video/avi");
		loadMimeType (".avs", "video/avs-video");
		loadMimeType (".bm", "image/bmp");
		loadMimeType (".bmp", "image/bmp");
		loadMimeType (".c", "text/x-c");
		loadMimeType (".c++", "text/x-c");
		loadMimeType (".cc", "text/x-c");
		loadMimeType (".conf", "text/plain");
		loadMimeType (".config", "text/xml");
		loadMimeType (".cpp", "text/x-c");
		loadMimeType (".css", "text/css");
		loadMimeType (".cxx", "text/plain");
		loadMimeType (".def", "text/plain");
		loadMimeType (".dif", "video/x-dv");
		loadMimeType (".dl", "video/x-dl");
		loadMimeType (".dv", "video/x-dv");
		loadMimeType (".dwf", "model/vnd.dwf");
		loadMimeType (".el", "text/x-script.elisp");
		loadMimeType (".etx", "text/x-setext");
		loadMimeType (".f", "text/x-fortran");
		loadMimeType (".fif", "image/fif");
		loadMimeType (".fli", "video/fli");
		loadMimeType (".flo", "image/florian");
		loadMimeType (".for", "text/x-fortran");
		loadMimeType (".fpx", "image/vnd.fpx");
		loadMimeType (".frl", "application/freeloader");
		loadMimeType (".g", "text/plain");
		loadMimeType (".gif", "image/gif");
		loadMimeType (".gtar", "application/x-gtar");
		loadMimeType (".gz", "application/x-gzip");
		loadMimeType (".gzip", "application/x-gzip");
		loadMimeType (".h", "text/plain");
		loadMimeType (".hh", "text/x-h");
		loadMimeType (".htc", "text/x-component");
		loadMimeType (".htm", "text/html");
		loadMimeType (".html", "text/html");
		loadMimeType (".htmls", "text/html");
		loadMimeType (".htt", "text/webviewhtml");
		loadMimeType (".htx", "text/html");
		loadMimeType (".ico", "image/x-icon");
		loadMimeType (".idc", "text/plain");
		loadMimeType (".ief", "image/ief");
		loadMimeType (".iefs", "image/ief");
		loadMimeType (".isu", "video/x-isvideo");
		loadMimeType (".it", "audio/it");
		loadMimeType (".ivr", "i-world/i-vrml");
		loadMimeType (".jav", "text/x-java-source");
		loadMimeType (".java", "text/x-java-source");
		loadMimeType (".jfif", "image/jpeg");
		loadMimeType (".jpe", "image/jpeg");
		loadMimeType (".jpeg", "image/jpeg");
		loadMimeType (".jpg", "image/jpeg");
		loadMimeType (".jps", "image/x-jps");
		loadMimeType (".js", "text/javascript");
		loadMimeType (".json", "text/json");
		loadMimeType (".kar", "music/x-karaoke");
		loadMimeType (".latex", "application/x-latex");
		loadMimeType (".list", "text/plain");
		loadMimeType (".log", "text/plain");
		loadMimeType (".lst", "text/plain");
		loadMimeType (".lsx", "text/x-la-asf");
		loadMimeType (".ltx", "application/x-latex");
		loadMimeType (".lzh", "application/octet-stream");
		loadMimeType (".lzx", "application/lzx");
		loadMimeType (".m", "text/plain");
		loadMimeType (".m1v", "video/mpeg");
		loadMimeType (".m2a", "audio/mpeg");
		loadMimeType (".m2v", "video/mpeg");
		loadMimeType (".m3u", "audio/x-mpequrl");
		loadMimeType (".man", "application/x-troff-man");
		loadMimeType (".map", "application/x-navimap");
		loadMimeType (".mar", "text/plain");
		loadMimeType (".mcf", "image/vasa");
		loadMimeType (".mid", "application/x-midi");
		loadMimeType (".midi", "application/x-midi");
		loadMimeType (".mm", "application/base64");
		loadMimeType (".mme", "application/base64");
		loadMimeType (".mod", "audio/mod");
		loadMimeType (".moov", "video/quicktime");
		loadMimeType (".mov", "video/quicktime");
		loadMimeType (".movie", "video/x-sgi-movie");
		loadMimeType (".mp2", "audio/mpeg");
		loadMimeType (".mp3", "audio/mpeg3");
		loadMimeType (".mpa", "audio/mpeg");
		loadMimeType (".mpc", "application/x-project");
		loadMimeType (".mpe", "video/mpeg");
		loadMimeType (".mpeg", "video/mpeg");
		loadMimeType (".mpg", "audio/mpeg");
		loadMimeType (".mpga", "audio/mpeg");
		loadMimeType (".mpp", "application/vnd.ms-project");
		loadMimeType (".mpt", "application/x-project");
		loadMimeType (".mpv", "application/x-project");
		loadMimeType (".mpx", "application/x-project");
		loadMimeType (".mrc", "application/marc");
		loadMimeType (".ms", "application/x-troff-ms");
		loadMimeType (".mv", "video/x-sgi-movie");
		loadMimeType (".my", "audio/make");
		loadMimeType (".mzz", "application/x-vnd.audioexplosion.mzz");
		loadMimeType (".nap", "image/naplps");
		loadMimeType (".naplps", "image/naplps");
		loadMimeType (".nc", "application/x-netcdf");
		loadMimeType (".ncm", "application/vnd.nokia.configuration-message");
		loadMimeType (".nif", "image/x-niff");
		loadMimeType (".niff", "image/x-niff");
		loadMimeType (".nix", "application/x-mix-transfer");
		loadMimeType (".nsc", "application/x-conference");
		loadMimeType (".nvd", "application/x-navidoc");
		loadMimeType (".o", "application/octet-stream");
		loadMimeType (".oda", "application/oda");
		loadMimeType (".omc", "application/x-omc");
		loadMimeType (".omcd", "application/x-omcdatamaker");
		loadMimeType (".omcr", "application/x-omcregerator");
		loadMimeType (".p", "text/x-pascal");
		loadMimeType (".p10", "application/pkcs10");
		loadMimeType (".p12", "application/pkcs-12");
		loadMimeType (".p7a", "application/x-pkcs7-signature");
		loadMimeType (".p7c", "application/pkcs7-mime");
		loadMimeType (".p7m", "application/pkcs7-mime");
		loadMimeType (".p7r", "application/x-pkcs7-certreqresp");
		loadMimeType (".p7s", "application/pkcs7-signature");
		loadMimeType (".part", "application/pro_eng");
		loadMimeType (".pas", "text/pascal");
		loadMimeType (".pbm", "image/x-portable-bitmap");
		loadMimeType (".pcl", "application/vnd.hp-pcl");
		loadMimeType (".pct", "image/x-pict");
		loadMimeType (".pcx", "image/x-pcx");
		loadMimeType (".pdb", "chemical/x-pdb");
		loadMimeType (".pdf", "application/pdf");
		loadMimeType (".pfunk", "audio/make");
		loadMimeType (".pgm", "image/x-portable-graymap");
		loadMimeType (".pic", "image/pict");
		loadMimeType (".pict", "image/pict");
		loadMimeType (".pkg", "application/x-newton-compatible-pkg");
		loadMimeType (".pko", "application/vnd.ms-pki.pko");
		loadMimeType (".pl", "text/plain");
		loadMimeType (".plx", "application/x-pixclscript");
		loadMimeType (".pm", "image/x-xpixmap");
		loadMimeType (".pm4", "application/x-pagemaker");
		loadMimeType (".pm5", "application/x-pagemaker");
		loadMimeType (".png", "image/png");
		loadMimeType (".pnm", "application/x-portable-anymap");
		loadMimeType (".pot", "application/mspowerpoint");
		loadMimeType (".pov", "model/x-pov");
		loadMimeType (".ppa", "application/vnd.ms-powerpoint");
		loadMimeType (".ppm", "image/x-portable-pixmap");
		loadMimeType (".pps", "application/mspowerpoint");
		loadMimeType (".ppt", "application/mspowerpoint");
		loadMimeType (".ppz", "application/mspowerpoint");
		loadMimeType (".pre", "application/x-freelance");
		loadMimeType (".prt", "application/pro_eng");
		loadMimeType (".ps", "application/postscript");
		loadMimeType (".psd", "application/octet-stream");
		loadMimeType (".pvu", "paleovu/x-pv");
		loadMimeType (".pwz", "application/vnd.ms-powerpoint");
		loadMimeType (".py", "text/x-script.phyton");
		loadMimeType (".pyc", "applicaiton/x-bytecode.python");
		loadMimeType (".qcp", "audio/vnd.qcelp");
		loadMimeType (".qd3", "x-world/x-3dmf");
		loadMimeType (".qd3d", "x-world/x-3dmf");
		loadMimeType (".qif", "image/x-quicktime");
		loadMimeType (".qt", "video/quicktime");
		loadMimeType (".qtc", "video/x-qtc");
		loadMimeType (".qti", "image/x-quicktime");
		loadMimeType (".qtif", "image/x-quicktime");
		loadMimeType (".ra", "audio/x-pn-realaudio");
		loadMimeType (".ram", "audio/x-pn-realaudio");
		loadMimeType (".ras", "application/x-cmu-raster");
		loadMimeType (".rast", "image/cmu-raster");
		loadMimeType (".rexx", "text/x-script.rexx");
		loadMimeType (".rf", "image/vnd.rn-realflash");
		loadMimeType (".rgb", "image/x-rgb");
		loadMimeType (".rm", "application/vnd.rn-realmedia");
		loadMimeType (".rmi", "audio/mid");
		loadMimeType (".rmm", "audio/x-pn-realaudio");
		loadMimeType (".rmp", "audio/x-pn-realaudio");
		loadMimeType (".rng", "application/ringing-tones");
		loadMimeType (".rnx", "application/vnd.rn-realplayer");
		loadMimeType (".roff", "application/x-troff");
		loadMimeType (".rp", "image/vnd.rn-realpix");
		loadMimeType (".rpm", "audio/x-pn-realaudio-plugin");
		loadMimeType (".rt", "text/richtext");
		loadMimeType (".rtf", "application/rtf");
		loadMimeType (".rtx", "application/rtf");
		loadMimeType (".rv", "video/vnd.rn-realvideo");
		loadMimeType (".s", "text/x-asm");
		loadMimeType (".s3m", "audio/s3m");
		loadMimeType (".saveme", "application/octet-stream");
		loadMimeType (".sbk", "application/x-tbook");
		loadMimeType (".scm", "application/x-lotusscreencam");
		loadMimeType (".sdml", "text/plain");
		loadMimeType (".sdp", "application/sdp");
		loadMimeType (".sdr", "application/sounder");
		loadMimeType (".sea", "application/sea");
		loadMimeType (".set", "application/set");
		loadMimeType (".sgm", "text/sgml");
		loadMimeType (".sgml", "text/sgml");
		loadMimeType (".sh", "application/x-bsh");
		loadMimeType (".shar", "application/x-bsh");
		loadMimeType (".shtml", "text/html");
		loadMimeType (".sid", "audio/x-psid");
		loadMimeType (".sit", "application/x-sit");
		loadMimeType (".skd", "application/x-koan");
		loadMimeType (".skm", "application/x-koan");
		loadMimeType (".skp", "application/x-koan");
		loadMimeType (".skt", "application/x-koan");
		loadMimeType (".sl", "application/x-seelogo");
		loadMimeType (".smi", "application/smil");
		loadMimeType (".smil", "application/smil");
		loadMimeType (".snd", "audio/basic");
		loadMimeType (".sol", "application/solids");
		loadMimeType (".spc", "application/x-pkcs7-certificates");
		loadMimeType (".spl", "application/futuresplash");
		loadMimeType (".spr", "application/x-sprite");
		loadMimeType (".sprite", "application/x-sprite");
		loadMimeType (".src", "application/x-wais-source");
		loadMimeType (".ssi", "text/x-server-parsed-html");
		loadMimeType (".ssm", "application/streamingmedia");
		loadMimeType (".sst", "application/vnd.ms-pki.certstore");
		loadMimeType (".step", "application/step");
		loadMimeType (".stl", "application/sla");
		loadMimeType (".stp", "application/step");
		loadMimeType (".sv4cpio", "application/x-sv4cpio");
		loadMimeType (".sv4crc", "application/x-sv4crc");
		loadMimeType (".svf", "image/vnd.dwg");
		loadMimeType (".svr", "application/x-world");
		loadMimeType (".swf", "application/x-shockwave-flash");
		loadMimeType (".t", "application/x-troff");
		loadMimeType (".talk", "text/x-speech");
		loadMimeType (".tar", "application/x-tar");
		loadMimeType (".tbk", "application/toolbook");
		loadMimeType (".tcl", "application/x-tcl");
		loadMimeType (".tcsh", "text/x-script.tcsh");
		loadMimeType (".tex", "application/x-tex");
		loadMimeType (".texi", "application/x-texinfo");
		loadMimeType (".texinfo", "application/x-texinfo");
		loadMimeType (".text", "application/plain");
		loadMimeType (".tgz", "application/gnutar");
		loadMimeType (".tif", "image/tiff");
		loadMimeType (".tiff", "image/tiff");
		loadMimeType (".tr", "application/x-troff");
		loadMimeType (".tsi", "audio/tsp-audio");
		loadMimeType (".tsp", "application/dsptype");
		loadMimeType (".tsv", "text/tab-separated-values");
		loadMimeType (".turbot", "image/florian");
		loadMimeType (".txt", "text/plain");
		loadMimeType (".uil", "text/x-uil");
		loadMimeType (".uni", "text/uri-list");
		loadMimeType (".unis", "text/uri-list");
		loadMimeType (".unv", "application/i-deas");
		loadMimeType (".uri", "text/uri-list");
		loadMimeType (".uris", "text/uri-list");
		loadMimeType (".ustar", "application/x-ustar");
		loadMimeType (".uu", "application/octet-stream");
		loadMimeType (".uue", "text/x-uuencode");
		loadMimeType (".vcd", "application/x-cdlink");
		loadMimeType (".vcs", "text/x-vcalendar");
		loadMimeType (".vda", "application/vda");
		loadMimeType (".vdo", "video/vdo");
		loadMimeType (".vew", "application/groupwise");
		loadMimeType (".viv", "video/vivo");
		loadMimeType (".vivo", "video/vivo");
		loadMimeType (".vmd", "application/vocaltec-media-desc");
		loadMimeType (".vmf", "application/vocaltec-media-file");
		loadMimeType (".voc", "audio/voc");
		loadMimeType (".vos", "video/vosaic");
		loadMimeType (".vox", "audio/voxware");
		loadMimeType (".vqe", "audio/x-twinvq-plugin");
		loadMimeType (".vqf", "audio/x-twinvq");
		loadMimeType (".vql", "audio/x-twinvq-plugin");
		loadMimeType (".vrml", "application/x-vrml");
		loadMimeType (".vrt", "x-world/x-vrt");
		loadMimeType (".vsd", "application/x-visio");
		loadMimeType (".vst", "application/x-visio");
		loadMimeType (".vsw", "application/x-visio");
		loadMimeType (".w60", "application/wordperfect6.0");
		loadMimeType (".w61", "application/wordperfect6.1");
		loadMimeType (".w6w", "application/msword");
		loadMimeType (".wav", "audio/wav");
		loadMimeType (".wb1", "application/x-qpro");
		loadMimeType (".wbmp", "image/vnd.wap.wbmp");
		loadMimeType (".web", "application/vnd.xara");
		loadMimeType (".wiz", "application/msword");
		loadMimeType (".wk1", "application/x-123");
		loadMimeType (".wmf", "windows/metafile");
		loadMimeType (".wml", "text/vnd.wap.wml");
		loadMimeType (".wmlc", "application/vnd.wap.wmlc");
		loadMimeType (".wmls", "text/vnd.wap.wmlscript");
		loadMimeType (".wmlsc", "application/vnd.wap.wmlscriptc");
		loadMimeType (".word", "application/msword");
		loadMimeType (".wp", "application/wordperfect");
		loadMimeType (".wp5", "application/wordperfect");
		loadMimeType (".wp6", "application/wordperfect");
		loadMimeType (".wpd", "application/wordperfect");
		loadMimeType (".wq1", "application/x-lotus");
		loadMimeType (".wri", "application/mswrite");
		loadMimeType (".wrl", "application/x-world");
		loadMimeType (".wrz", "model/vrml");
		loadMimeType (".wsc", "text/scriplet");
		loadMimeType (".wsrc", "application/x-wais-source");
		loadMimeType (".wtk", "application/x-wintalk");
		loadMimeType (".xbm", "image/x-xbitmap");
		loadMimeType (".xdr", "video/x-amt-demorun");
		loadMimeType (".xgz", "xgl/drawing");
		loadMimeType (".xif", "image/vnd.xiff");
		loadMimeType (".xl", "application/excel");
		loadMimeType (".xla", "application/excel");
		loadMimeType (".xlb", "application/excel");
		loadMimeType (".xlc", "application/excel");
		loadMimeType (".xld", "application/excel");
		loadMimeType (".xlk", "application/excel");
		loadMimeType (".xll", "application/excel");
		loadMimeType (".xlm", "application/excel");
		loadMimeType (".xls", "application/excel");
		loadMimeType (".xlt", "application/excel");
		loadMimeType (".xlv", "application/excel");
		loadMimeType (".xlw", "application/excel");
		loadMimeType (".xm", "audio/xm");
		loadMimeType (".xml", "text/xml");
		loadMimeType (".xmz", "xgl/movie");
		loadMimeType (".xpix", "application/x-vnd.ls-xpix");
		loadMimeType (".xpm", "image/xpm");
		loadMimeType (".x-png", "image/png");
		loadMimeType (".xsr", "video/x-amt-showrun");
		loadMimeType (".xwd", "image/x-xwd");
		loadMimeType (".xyz", "chemical/x-pdb");
		loadMimeType (".z", "application/x-compress");
		loadMimeType (".zip", "application/zip");
		loadMimeType (".zoo", "application/octet-stream");
		loadMimeType (".zsh", "text/x-script.zsh");
	}


} // namespace ahttp

