/*
This file is part of [ahttpserver] project. 

Author: Artem Kustikov (kustikoff[at]tut.by)

This code is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any
damages arising from the use of this code.

Permission is granted to anyone to use this code for any
purpose, including commercial applications, and to alter it and
redistribute it freely, subject to the following restrictions:

1. The origin of this code must not be misrepresented; you must
not claim that you wrote the original code. If you use this
code in a product, an acknowledgment in the product documentation
would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and
must not be misrepresented as being the original code.

3. This notice may not be removed or altered from any source
distribution.
*/

#if defined (WIN32)
#	include <signal.h>
#elif defined (__GNUC__)
#	include <sys/signal.h>
#endif  //__GNUC__

#include <boost/program_options.hpp>
#include <boost/filesystem.hpp>

namespace po = boost::program_options;
namespace fs = boost::filesystem;

#include "ahttplib.hpp"

//////////////////////////////////////////////////////////////////////////

ahttp::HttpServerSettings globalSettings;
ahttp::HttpServer httpServer;
aconnect::ConsoleLogger logger(aconnect::Log::Info); // FakeLogger can be used

void processSignal (int sig)
{
	httpServer.Engine.stop(false);
	
	std::cout << "Server stopped" << std::endl;
}

int main (int ac, char* av[]) 
{
	signal (SIGINT, processSignal);
	signal (SIGABRT, processSignal);
	#ifdef WIN32
		signal (SIGBREAK, processSignal);
	#endif

	using std::cout;
	using std::endl;

	aconnect::port_type port;
	fs::path rootDir;

	po::options_description desc("Allowed options");

	desc.add_options()
		("help", "Produce help message")
		("port", po::value<aconnect::port_type>(&port)->default_value(ahttp::port::http), 
			"Port to listen")
			("root", po::value<fs::path>(&rootDir)->default_value(fs::initial_path()), 
			"Folder to share files through HTTP")
	;

	po::variables_map vm;
	po::store(po::parse_command_line(ac, av, desc), vm);
	po::notify(vm); 


	if (vm.count("help")) {
		cout << desc << "\n";
		return 1;
	}


	cout << "Port: " << port << endl;
	cout << "HTTP root: " << rootDir << endl;

	try
	{
		// get it working
		aconnect::Initializer socketsListInit; 
				
		globalSettings.setLogger (&logger);
		globalSettings.setPort (port);
		globalSettings.setupDefaultMimeTypes ();
		
		ahttp::DirectorySettings rootDirInfo;
		rootDirInfo.realPath = rootDir.directory_string();
		
		globalSettings.setupRootDirectory (rootDirInfo, 0);
		
		httpServer.init (&globalSettings);
		httpServer.Engine.start (true);
	
	} catch (std::exception &ex)  {
		cout << "Server startup failed - exception [" << typeid(ex).name() << "]: " << ex.what() << endl;

	} catch (...)  {
		cout << ("Server startup failed - unknown exception caught") << endl;
	}
}