/*
This file is part of [ahttpserver] project. 

Author: Artem Kustikov (kustikoff[at]tut.by)


This code is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any
damages arising from the use of this code.

Permission is granted to anyone to use this code for any
purpose, including commercial applications, and to alter it and
redistribute it freely, subject to the following restrictions:

1. The origin of this code must not be misrepresented; you must
not claim that you wrote the original code. If you use this
code in a product, an acknowledgment in the product documentation
would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and
must not be misrepresented as being the original code.

3. This notice may not be removed or altered from any source
distribution.
*/

#ifdef WIN32

#include "aconnect/config.hpp"

#include "vcclr.h"

#include "ahttplib.hpp"
#include "aconnect/util.hpp"
#include "aconnect/util.string.hpp"
#include "aconnect/util.file.hpp"
#include "aconnect/util.time.hpp"


#include <boost/algorithm/string.hpp>
#include <boost/thread.hpp>

namespace algo = boost::algorithm;

using namespace System;
using namespace System::IO;
using namespace System::Web;
using namespace System::Web::Hosting;
using namespace System::Threading;
using namespace System::Collections::Generic;
using namespace System::Reflection;
using namespace System::Runtime::Remoting;

// using namespace AHttp;

namespace Globals
{
	// constants
	const aconnect::string Param_InitRoot = "init-root";
	const aconnect::string Param_LoadApplications = "load-applications";
    const aconnect::string Param_RecyclingTimeout = "recycling-timeout";

	const aconnect::string Param_AspRecycleByRequestsCount = "asp-recycle-by-requests-count";
	const aconnect::string Param_AspRecycleByUptime = "asp-recycle-by-uptime";
    const aconnect::string Param_AspReloadUrl = "asp-reload-url";
	
	// objects
	boost::mutex LoadMutex;
	
	boost::thread *RecycleThread;
	boost::condition DestroyCondition;
	
	static ahttp::HttpServer *HttpServer = NULL;

	// params
	int RecyclingTimeout = 30; //sec
}

#include "../shared/managed_strings_conversions.inl"
#include "managed.inl"

HANDLER_EXPORT bool initPlugin  (const aconnect::str2str_map& params, 
								  int handlerIndex,
								  ahttp::HttpServer *server);
HANDLER_EXPORT void destroyPlugin  ();
HANDLER_EXPORT bool processHandlerRequest (ahttp::HttpContext& context, int handlerIndex);


void processException (aconnect::string_constptr prefix);
void processRecycling (int);

//////////////////////////////////////////////////////////////////////////

#pragma unmanaged

BOOL APIENTRY DllMain( HMODULE hModule,
					  DWORD  ul_reason_for_call,
					  LPVOID lpReserved
					  )
{
	if (ul_reason_for_call == DLL_PROCESS_ATTACH)
	{
		// Don't need to be called for new threads
		DisableThreadLibraryCalls ((HMODULE) hModule);
	}
	else if (ul_reason_for_call == DLL_PROCESS_DETACH)
	{

	}
	return TRUE;
}


#pragma managed



/* 
*	Handler initialization function - return true if initialization performed sucessfully
*/

HANDLER_EXPORT bool initPlugin (const aconnect::str2str_map& params, 
								  int handlerIndex,
								  ahttp::HttpServer *server)
{
	assert (server);
	assert (server->Log());
	assert (server->GlobalSettings());
	assert (server->GlobalSettings()->logger());
	
	boost::mutex::scoped_lock lock(Globals::LoadMutex);
	
	if (Globals::HttpServer) {
		server->Log()->error ("Handler already initialized - ASP.NET handler cannot be loaded twice");
		return false;
	}

	Globals::HttpServer = server;
    
    Globals::RecyclingTimeout = aconnect::util::getItemFromMap (params, Globals::Param_RecyclingTimeout, Globals::RecyclingTimeout);
    
	try
	{
		if (aconnect::util::getItemFromMapBool (params, Globals::Param_InitRoot, true)) 
		{
			const ahttp::DirectorySettings& root = server->GlobalSettings()->getRootDirSettings ();
			AHttpInternal::Application::Server->InitHost (root, true);
		}
		
		// process "load-applications" parameter
		aconnect::string loadApps = aconnect::util::getItemFromMap (params, Globals::Param_LoadApplications);
		if (!loadApps.empty()) 
		{
			aconnect::SimpleTokenizer tokens(loadApps, ";");
			aconnect::SimpleTokenizer::iterator tok_iter;
		
			for (tok_iter = tokens.begin(); tok_iter != tokens.end(); ++tok_iter)
			{
				aconnect::string token = tok_iter.current_token();
				algo::trim (token);

				if (token.empty())
					continue;

				const ahttp::DirectorySettings* dir = server->GlobalSettings()->getDirSettingsByName (token);
				if (0 == dir) {
					server->Log()->warn ("Directory settings not found by name \"%s\"", token.c_str());
				} else {
					AHttpInternal::Application::Server->InitHost (*dir, true);
				}
			}
		}
		
	} catch (std::exception const &ex)  {
		server->Log()->error ("ASPNET, Initialization: exception caught (%s): %s", 
			typeid(ex).name(), ex.what());
	
	} catch (System::Exception^ ex) {
		server->Log()->error ("ASPNET, Initialization: managed exception caught: %s", 
			AHttp::getUnmanagedString(ex->ToString()).c_str());
		return false;

	} catch (...) {
		processException("Initialization");
		return false;
	}

	try	{
		
		Globals::RecycleThread = new boost::thread (processRecycling, 0);
		
	} catch (...) {
		processException("Recycling thread start");
		return false;
	}

	return true;
}


HANDLER_EXPORT void destroyPlugin  ()
{
	if (0 == Globals::HttpServer)
		return; // already cleaned

	boost::mutex::scoped_lock lock(Globals::LoadMutex);
	Globals::DestroyCondition.notify_one();

	for each ( KeyValuePair <int, AHttp::AspNetHost^> pair in AHttpInternal::Application::Server->Hosts )
		pair.Value->Stop();

	AHttpInternal::Application::Server->Hosts->Clear();
	
	delete Globals::RecycleThread;
	
	Globals::HttpServer = 0;
	Globals::RecycleThread = 0;
}


/* 
*	Main request processing function - return false if request should
*	be processed by other handlers or by server (true: request completed)
*/
HANDLER_EXPORT bool processHandlerRequest (ahttp::HttpContext& context, 
										   int handlerIndex)
{
	using namespace ahttp;
	using namespace aconnect;
	
	try
	{
		AHttp::AspNetHost^ host = nullptr;

		try
		{
			System::Threading::Monitor::Enter (AHttpInternal::Application::Server);
			
			if ( AHttpInternal::Application::Server->Hosts->ContainsKey(context.CurrentDirectoryInfo->number) ) 
			{
				try {
					host = AHttpInternal::Application::Server->Hosts[context.CurrentDirectoryInfo->number];
					host->GetDomain();

				} catch (AppDomainUnloadedException^) {
					context.Log->warn ("ASPNET: application domain was unloaded, directory: %s",
						context.CurrentDirectoryInfo->name.c_str());
					
                    AHttpInternal::Application::Server->Hosts->Remove(context.CurrentDirectoryInfo->number);
					host = nullptr;
				}
			}
		} 
		finally
		{
			System::Threading::Monitor::Exit (AHttpInternal::Application::Server);
		}
        			
        if (host == nullptr)
            host = AHttpInternal::Application::Server->InitHost (*context.CurrentDirectoryInfo, false); 
		
		// give request to HttpRuntime
		host->ProcessRequest((void*) &context);
		
		
		// update Content-Type if not set
		if (!context.isClosed())
			context.setHtmlResponse();
	
	} catch (System::Exception^ ex) {
		Globals::HttpServer->Log()->error ("ASPNET, request processing: managed exception caught: %s", 
			AHttp::getUnmanagedString(ex->ToString()).c_str());

		context.Server->processServerError(context, ahttp::HttpStatus::InternalServerError,
			"ASPNET Exception caught: %s", 
			AHttp::getUnmanagedString(ex->ToString()).c_str());

		Globals::HttpServer->Log()->error ("ASPNET, request processing: managed exception caught: %s", 
			AHttp::getUnmanagedString(ex->ToString()).c_str());

	
	} catch (...) {
		processException("Processing request");
	}


	return true;
}

void processException (aconnect::string_constptr prefix)
{
	try
	{
		throw;

	} catch (std::exception const &ex)  {
		Globals::HttpServer->Log()->error ("ASPNET, %s: exception caught (%s): %s", 
			prefix, typeid(ex).name(), ex.what());
	
	} catch (...)  {
		Globals::HttpServer->Log()->error ("ASPNET, %s: unknown exception caught", prefix);
		
	}
}

void processRecycling (int)
{
	do 
	{
		boost::mutex::scoped_lock lock(Globals::LoadMutex);
		Globals::DestroyCondition.timed_wait (lock, aconnect::util::createTimePeriod (Globals::RecyclingTimeout));

		if (0 == Globals::HttpServer || Globals::HttpServer->isStopped())
			break;

		System::Threading::Monitor::Enter (AHttpInternal::Application::Server);

		Dictionary<int, AHttp::AspNetHost^> ^hostsCopy = gcnew Dictionary<int, AHttp::AspNetHost^> (AHttpInternal::Application::Server->Hosts->Count);

		for each ( KeyValuePair <int, AHttp::AspNetHost^> pair in AHttpInternal::Application::Server->Hosts )
			hostsCopy->Add (pair.Key, pair.Value);

		System::Threading::Monitor::Exit (AHttpInternal::Application::Server);
		
		for each ( KeyValuePair <int, AHttp::AspNetHost^> pair in hostsCopy )
		{
			const ahttp::DirectorySettings* settings = Globals::HttpServer->GlobalSettings()->getDirSettings(pair.Key);
			assert (settings);

			try 
			{
				AppDomain^ domain = pair.Value->GetDomain();
				
				if (Globals::HttpServer->Log()->isDebugEnabled())
					Globals::HttpServer->Log()->debug ("ASPNET: recycling - application domain checked, directory: %s, name: %s, requests count: %d, memory: %d",
						AHttp::getUnmanagedString(domain->BaseDirectory).c_str(), 
						AHttp::getUnmanagedString(domain->FriendlyName).c_str(),
						pair.Value->GetRequestsCount(),
						System::GC::GetTotalMemory(false));

				
				bool stopped = false;
				
				long limitRecycleRequestCount = aconnect::util::getItemFromMap (settings->customParams, Globals::Param_AspRecycleByRequestsCount, -1);
				long currentRequestCount = pair.Value->GetRequestsCount();
				
				if (limitRecycleRequestCount > 0) {
					if (currentRequestCount >= limitRecycleRequestCount) {
						pair.Value->Stop();
						stopped = true;

						Globals::HttpServer->Log()->info ("ASPNET: application reloaded by requests count, name: %s, req. count: %d",
							settings->name.c_str(),
							currentRequestCount);
					}
				}

				if (!stopped) {
					long limitUptimeTimeout = aconnect::util::getItemFromMap (settings->customParams, Globals::Param_AspRecycleByUptime, -1);
				
					if (limitUptimeTimeout > 0) {
						
						__int64 uptimeTicks = DateTime::UtcNow.Ticks - pair.Value->StartTimeStamp->Ticks;

						if (TimeSpan::FromTicks(uptimeTicks) >= TimeSpan::FromMinutes(limitUptimeTimeout)) {
							pair.Value->Stop();
							stopped = true;

							Globals::HttpServer->Log()->info ("ASPNET: application reloaded by uptime, name: %s",
								settings->name.c_str());
						}
					}	
				}
				

			} catch (AppDomainUnloadedException^) {
				Globals::HttpServer->Log()->debug ("ASPNET: recycling - found unloaded application domain, directory: %d",
						pair.Key);

				// remove previosly loaded host
				System::Threading::Monitor::Enter (AHttpInternal::Application::Server);
				AHttpInternal::Application::Server->Hosts->Remove(pair.Key);
				System::Threading::Monitor::Exit (AHttpInternal::Application::Server);

				AHttpInternal::Application::Server->InitHost (*settings, true); 
			}
		}

	} while (Globals::HttpServer && !Globals::HttpServer->isStopped());
	
}


////////////////////////////////////////////////////////////////////////////////

#endif // WIN32

