/*
This file is part of [ahttpserver] project. 

Author: Artem Kustikov (kustikoff[at]tut.by)


This code is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any
damages arising from the use of this code.

Permission is granted to anyone to use this code for any
purpose, including commercial applications, and to alter it and
redistribute it freely, subject to the following restrictions:

1. The origin of this code must not be misrepresented; you must
not claim that you wrote the original code. If you use this
code in a product, an acknowledgment in the product documentation
would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and
must not be misrepresented as being the original code.

3. This notice may not be removed or altered from any source
distribution.
*/


#include "ahttplib.hpp"
#include "aconnect/util.hpp"
#include "aconnect/util.string.hpp"
#include "aconnect/util.file.hpp"
#include "aconnect/util.guards.hpp"

#include <boost/thread.hpp>
#include <boost/thread/condition.hpp>
#include <boost/filesystem.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/scoped_array.hpp>
#include <boost/algorithm/string.hpp>

#include "fastcgi/fastcgi.hpp"
#include "fastcgi/fastcgi_util.hpp"
#include "fastcgi/fastcgi_application_manager.hpp"

namespace fs = boost::filesystem;
namespace algo = boost::algorithm;

typedef std::list< std::pair <int, bool> > process_list; // implement objects pool

//////////////////////////////////////////////////////////////////////////

bool findFreeApplication (const std::pair <int, bool>& p) {
	return p.second;
}

struct findApplicationById : public std::unary_function<std::pair <int, bool>, bool>
{
	int id;

	findApplicationById (int id_) : id (id_) {
	}
	
	bool operator() (const std::pair <int, bool>& p) const {
		return p.first == id;
	}
};

//////////////////////////////////////////////////////////////////////////

struct FcgiRecord
{
	fs::path	path;
	std::string commandLine;

	int			maxInstancesCount;
	int			handlerIndex;
	
	//	properties
	aconnect::port_type tcpListenerPort;
	bool				checkThatFileExists;
	bool				useGetValues;
	bool				useWebServerAddrs;
	int					backlog;

	//	FastCGI
	fastcgi::ConnectionMode mode;
	process_list applicationsList;
	boost::mutex processMutex;
	boost::condition processCondition;
	
	int sockectSelectTimeout; // sec
	int sockectReadWriteTimeout; // sec

	fastcgi::FcgiApplicationInfo appInfo;
	int lastRequestId;
	
	FcgiRecord():
		maxInstancesCount (10),
		handlerIndex (0),
		tcpListenerPort (0),
		checkThatFileExists (true),
		useGetValues (false),
		useWebServerAddrs (true),
		backlog (5),

		mode (fastcgi::ConnectionPipe),
		sockectSelectTimeout (fastcgi::os::DefaultSocketSelectTimeout),		// sec
		sockectReadWriteTimeout (fastcgi::os::DefaultReadWriteTimeout),	// sec

		lastRequestId (1)
	{ 

	}
};



namespace Globals
{

	const aconnect::string Param_Engine	= "engine";
	const aconnect::string Param_CommandLine	= "command-line";
	const aconnect::string Param_Mode	= "mode"; // possible values: "pipe", "tcp", default: "pipe"
	const aconnect::string Param_TcpListenerPort	= "tcp-listener-port";
	const aconnect::string Param_InstancesCount	= "instances-count";
	const aconnect::string Param_CheckThatFileExists	= "check-file-exists";
	const aconnect::string Param_UseGetValues = "use-get-values-info";
	const aconnect::string Param_UseWebServerAddrs = "use-web-server-addrs";
	const aconnect::string Param_Backlog = "backlog";
	const aconnect::string Param_ReadWriteTimeout	= "read-write-timeout";


	const aconnect::string Param_ModePipe	= "pipe";
	const aconnect::string Param_ModeTcpSocket	= "tcp";
	
	// FastCGI
	const aconnect::string PipePrefix = "ahttp-";


	// globals
	typedef std::map<int, FcgiRecord*> engines_map;
	

	boost::mutex LoadMutex;
	engines_map RegisteredEnginesMap;
	fastcgi::ApplicationManager AppManager;

	static ahttp::HttpServer *Server = NULL;
	static ahttp::Logger *Log = NULL;

}

HANDLER_EXPORT bool initPlugin  (const aconnect::str2str_map& params, 
								  int handlerIndex,
								  ahttp::HttpServer *server);
HANDLER_EXPORT void destroyPlugin  ();
HANDLER_EXPORT bool processHandlerRequest (ahttp::HttpContext& context, int handlerIndex);

//////////////////////////////////////////////////////////////////////////

void runEngines (FcgiRecord* fcgiRecord);
void killFastCgiApplications ();
void processException (aconnect::string_constptr prefix);
int startFcgiApplication (FcgiRecord* fcgiRecord, int desiredIndex = -1, bool initiallyFree = true) throw (fastcgi::fastcgi_error);
void closeFcgiConnection (int appDescriptor, FcgiRecord* record);

//////////////////////////////////////////////////////////////////////////
//	
//	Windows related stuff

#if defined (WIN32)
BOOL APIENTRY DllMain( HMODULE hModule,
					  DWORD  ul_reason_for_call,
					  LPVOID lpReserved
					  )
{
	if (ul_reason_for_call == DLL_PROCESS_ATTACH) {
		// Don't need to be called for new threads
		::DisableThreadLibraryCalls ((HMODULE) hModule);
	}
	return TRUE;
}
#endif // WIN32

/* 
*	Handler initialization function - return true if initialization performed sucessfully
*/
HANDLER_EXPORT bool initPlugin  (const aconnect::str2str_map& params, 
								  int handlerIndex,
								  ahttp::HttpServer *server)
{
	assert (server);
	assert (server->GlobalSettings());
	assert (server->Log());

	boost::mutex::scoped_lock lock(Globals::LoadMutex);
	
	try {
		
		if (Globals::Server == NULL) 
		{
			Globals::Server = server;
			Globals::Log = server->Log();
			
			Globals::AppManager.setLog (Globals::Log);
			Globals::AppManager.initFastCgiEnvironment();
		}
	
		FcgiRecord* currentRecord = new FcgiRecord();
		std::auto_ptr <FcgiRecord> recordGuard(currentRecord);

		currentRecord->handlerIndex = handlerIndex;

		// load engine settings
		aconnect::str2str_map::const_iterator it = params.find(Globals::Param_Engine);
		if (it != params.end()) {
			currentRecord->path = fs::system_complete (fs::path(it->second, fs::native));
			
			if (!fs::exists (currentRecord->path)) {
				server->Log()->error ("Fast CGI engine not found: %s", 
					currentRecord->path.file_string().c_str() );
				return false;
			}

		} else {
			server->Log()->error ("Mandatory parameter '%s' is absent", 
				Globals::Param_Engine.c_str() );
			return false;
		}

		// load command line - used for scripting engines (Python)
		currentRecord->commandLine = aconnect::util::getItemFromMap(params, Globals::Param_CommandLine);


		// load mode
		it = params.find(Globals::Param_Mode);
		if (it != params.end()) {

			if (aconnect::util::equals (it->second, Globals::Param_ModeTcpSocket) ) 
			{
				currentRecord->mode = fastcgi::ConnectionTcpSocket;

				aconnect::str2str_map::const_iterator port = params.find(Globals::Param_TcpListenerPort);
				
				if (port == params.end()) {
					server->Log()->error ("Mandatory TCP connection parameter '%s' is absent", Globals::Param_TcpListenerPort.c_str() );
					return false;
				}

				currentRecord->tcpListenerPort = boost::lexical_cast<aconnect::port_type, aconnect::string> (port->second);
			}
		}

		currentRecord->maxInstancesCount = aconnect::util::getItemFromMap (params, Globals::Param_InstancesCount, currentRecord->maxInstancesCount);
		currentRecord->checkThatFileExists = aconnect::util::getItemFromMapBool (params, Globals::Param_CheckThatFileExists, currentRecord->checkThatFileExists);
		currentRecord->useGetValues = aconnect::util::getItemFromMapBool (params, Globals::Param_UseGetValues, currentRecord->useGetValues);
		currentRecord->useWebServerAddrs = aconnect::util::getItemFromMapBool (params, Globals::Param_UseWebServerAddrs, currentRecord->useWebServerAddrs);
		
		currentRecord->backlog = aconnect::util::getItemFromMap (params, Globals::Param_Backlog, currentRecord->backlog);
		currentRecord->sockectReadWriteTimeout = aconnect::util::getItemFromMap (params, Globals::Param_ReadWriteTimeout, currentRecord->sockectReadWriteTimeout);
		
		runEngines (currentRecord);
		
		Globals::RegisteredEnginesMap[handlerIndex] = currentRecord;
		
		recordGuard.release();
		return true;

	} catch (...) {
		
		killFastCgiApplications ();
		processException ("FASTCGI: initPlugin");
	}

	return false;
}


HANDLER_EXPORT void destroyPlugin  ()
{
	boost::mutex::scoped_lock lock(Globals::LoadMutex);

	if (NULL == Globals::Server)
		return; // already cleaned

	killFastCgiApplications ();

	Globals::engines_map::iterator iter = Globals::RegisteredEnginesMap.begin();
	for (; iter != Globals::RegisteredEnginesMap.end(); ++iter) 
	{
		delete iter->second;
		iter->second = NULL;
	}

	Globals::Server = NULL;
}
void killFastCgiApplications ()
{
	Globals::AppManager.terminateApplications();

	Globals::engines_map::iterator iter = Globals::RegisteredEnginesMap.begin();
	
	for (; iter != Globals::RegisteredEnginesMap.end(); ++iter) 
	{
		FcgiRecord *record = iter->second;
		record->applicationsList.clear();
	}
}

int startFcgiApplication (FcgiRecord* fcgiRecord, int desiredIndex, bool initiallyFree) throw (fastcgi::fastcgi_error)
{
	using namespace Globals;
	
	size_t regIndex = fcgiRecord->applicationsList.size();

	if (desiredIndex != -1) {
		process_list::iterator iter = std::find_if(fcgiRecord->applicationsList.begin(), 
			fcgiRecord->applicationsList.end(), 
			findApplicationById(desiredIndex) );
		
		assert (iter != fcgiRecord->applicationsList.end());

		// copy previous instance index
		regIndex = std::distance (fcgiRecord->applicationsList.begin(), iter);
	}

	std::string pipeName = PipePrefix 
		+ boost::lexical_cast<std::string>(fcgiRecord->handlerIndex)
		+ "-" 
		+ boost::lexical_cast<std::string>(regIndex);

	int id = AppManager.createLocalIpcFileDescriptor (fcgiRecord->mode, 
			fcgiRecord->sockectReadWriteTimeout,
			desiredIndex,
			pipeName.c_str(), 
			fcgiRecord->tcpListenerPort + regIndex,
			fcgiRecord->backlog);
	
	AppManager.spawnWorkerProcess (fcgiRecord->path.file_string().c_str(), id, 
		(fcgiRecord->commandLine.empty() ? 0 : fcgiRecord->commandLine.c_str()));

	AppManager.close (id);

	if (desiredIndex == -1) {
		fcgiRecord->applicationsList.push_back (std::make_pair(id, initiallyFree)); // free at creation
	} else if (desiredIndex != id) {
		throw aconnect::application_error ("Cannot start FastCGI application with desired instance Id");
	}

	return id;
}

void closeFcgiConnection (int appDescriptor, FcgiRecord* record) 
{
	boost::mutex::scoped_lock recordLock (record->processMutex);
	
	process_list::iterator appIter = record->applicationsList.begin();
	while (appIter != record->applicationsList.end()) {
		if (appIter->first == appDescriptor) {
	
			appIter->second = true;
			record->processCondition.notify_one();
			return;
		}

		++appIter;
	}

	assert (false && "Incorect beahvior - check closeFcgiConnection usage");
}


void runEngines (FcgiRecord* fcgiRecord)
{
	// prepare environment
	if (fcgiRecord->useWebServerAddrs) 
	{
		const aconnect::ip_addr_type &serverIp = Globals::Server->Engine.settings().ip;
		aconnect::string ipAddressStr = aconnect::util::isLocalhostIpAddress(serverIp) 
			? aconnect::util::formatIpAddr(aconnect::network::LocalhostAddress)
			: aconnect::util::formatIpAddr(serverIp);

#if defined (WIN32)
		if (0 == ::SetEnvironmentVariableA(fastcgi::FcgiWebServersEnvVariable, ipAddressStr.c_str()))
			throw fastcgi::fastcgi_error(::GetLastError(), "Setup 'FCGI_WEB_SERVER_ADDRS' environment variable failed.");
#else

#endif
	}

	// run 1 instance for to check availability
	int id = startFcgiApplication (fcgiRecord);
	
	// some strange error exists for 'pipe' connection on php-cgi (FCGI_GET_VALUES)
	// will work without it
	if (fcgiRecord->useGetValues) 
	{
		fastcgi::SocketDescriptor sd = Globals::AppManager.connectToFcgiApplication (id, fcgiRecord->sockectReadWriteTimeout);
		aconnect::util::scoped_param_guard<void(*)(fastcgi::SocketDescriptor), fastcgi::SocketDescriptor> socketGuard(fastcgi::ApplicationManager::closeDescriptor, sd);

		// send FCGI_GET_VALUES to test connection
		Globals::AppManager.loadFcgiAppInfo (sd, fcgiRecord->appInfo);
	}
}

void processException (aconnect::string_constptr prefix)
{
	try
	{
		throw;

	} catch (std::exception const &ex)  {
		if (Globals::Log) Globals::Log->error ("%s: exception caught (%s): %s", prefix, typeid(ex).name(), ex.what());
	} catch (...)  {
		if (Globals::Log) Globals::Log->error ("%s: unknown exception caught", prefix);
	}
}



void beginRequest (ahttp::HttpContext& context, 
				   const fastcgi::SocketDescriptor& sd, 
				   const int requestId);

/* 
*	Main request processing function - return false if request should
*	be processed by other handlers or by server (true: request completed)
*/
HANDLER_EXPORT bool processHandlerRequest (ahttp::HttpContext& context, 
										   int handlerIndex)
{
	using namespace ahttp;
	using namespace aconnect;
	using namespace fastcgi;
	

	Globals::engines_map::iterator iter = Globals::RegisteredEnginesMap.find (handlerIndex);
	if (iter == Globals::RegisteredEnginesMap.end()) {
		Globals::Log->error ("FASTCGI: Handler registration not found, index: %d", handlerIndex);
		context.Server->processServerError (context, HttpStatus::InternalServerError, "Handler registration not found");
		return true;
	}

	FcgiRecord* record = iter->second;

	if (record->applicationsList.size() == 0) {
		Globals::Log->error ("FASTCGI: no connected FastCGI applications found");
		context.Server->processServerError (context, HttpStatus::InternalServerError, "FASTCGI: no connected FastCGI applications found");
		return true;
	}

	if (record->checkThatFileExists) {
		if ( !util::fileExists (context.FileSystemPath.string()) ) {
			context.Server->processError404 (context);
			return true;
		}
	}

	if (!context.isClientConnected())
		return true;

	// FastCGI specific
	context.Items[strings::ServerVariables::GATEWAY_INTERFACE] = strings::GatewayInterfaceCgi1_1;

	try
	{
		int requestId = -1;
		int currentAppDescriptor =  -1;
		SocketDescriptor sd;

		boost::mutex::scoped_lock recordLock (record->processMutex);
		requestId = record->lastRequestId++;
		
		process_list::iterator appIter = std::find_if (record->applicationsList.begin(), 
			record->applicationsList.end(),
			findFreeApplication);

		if (appIter == record->applicationsList.end()) {
			
			if (record->applicationsList.size() < (size_t) record->maxInstancesCount ) {
				currentAppDescriptor = startFcgiApplication (record, -1, false);

			} else {
				bool waitRes = record->processCondition.timed_wait(recordLock, 
					aconnect::util::createTimePeriod (record->sockectReadWriteTimeout));

				if (waitRes ) {
					
					appIter = std::find_if (record->applicationsList.begin(), 
						record->applicationsList.end(),
						findFreeApplication);

					if (appIter != record->applicationsList.end()) {
						currentAppDescriptor = appIter->first;
						appIter->second = false;
					} else {
						waitRes = false;
					}
					
				} 
				
				if (!waitRes) {
					context.Server->processServerError(context, ahttp::HttpStatus::ServiceUnavailable, "FASTCGI: Application overload - please increase Max-Instances count." );
					return true;
				}
			}

		} else {
			currentAppDescriptor = appIter->first;
			appIter->second = false;
		}
		
		assert (std::find_if(record->applicationsList.begin(),
			record->applicationsList.end(),
			findApplicationById(currentAppDescriptor))->second == false);

		recordLock.unlock();

		aconnect::util::scoped_param_guard2<void(*)(int, FcgiRecord*), int, FcgiRecord*> connectionGuard(closeFcgiConnection, currentAppDescriptor, record);
		
		try {
			
			sd = Globals::AppManager.connectToFcgiApplication (currentAppDescriptor, record->sockectReadWriteTimeout);
			beginRequest(context, sd, requestId);
			
		} catch (fastcgi_error &err) {
			// FastCGI application unloaded - try start again (only once)
			if (err.code() == FcgiError::ClientConnectionFailed) {
			
				Globals::AppManager.freeConnection (currentAppDescriptor);
				currentAppDescriptor = startFcgiApplication (record, currentAppDescriptor, false);
				sd = Globals::AppManager.connectToFcgiApplication (currentAppDescriptor, record->sockectReadWriteTimeout);
	
				beginRequest(context, sd, requestId);
			
			} else {
				throw;
			}
		} 

		aconnect::util::scoped_param_guard<void(*)(SocketDescriptor), SocketDescriptor> socketGuard(ApplicationManager::closeDescriptor, sd);

		// Read response records
		InputStream input(sd, 2*MaxMessageContentLength, record->sockectReadWriteTimeout);
		Header responseHeader;
		bool headersLoaded = false;

		do
		{
			if (!input.readHeader (responseHeader))
				throw fastcgi_error(FcgiError::ProtocolError, "FASTCGI: StdOut/StdErr header corrupted.");

			if (responseHeader.version > FcgiVersion1)
				throw fastcgi_error(FcgiError::FastcgiInvalidVersion, "FASTCGI: unsupported FasCGI response version.");

			int outRequestId = (int)responseHeader.requestIdB0 + (int) (responseHeader.requestIdB1 << 8);
			
			if (outRequestId != requestId)
				throw fastcgi_error(FcgiError::ProtocolError, "FASTCGI: invalid request ID in response, %d instead of %d.", 
					outRequestId, requestId);

			if (responseHeader.type == FcgiMessage::EndRequest)
			{
				// read end request data
				EndRequestBody endRequestBody;

				if (input.readBlock ((fastcgi::byte_type*) &endRequestBody, sizeof(EndRequestBody)) != sizeof(EndRequestBody))
					throw fastcgi_error(FcgiError::ProtocolError, "FASTCGI: EndRequest response corrupted.");

				if (endRequestBody.protocolStatus != FcgiProtocolStatus::RequestComplete)
					Globals::Log->warn ("FASTCGI: EndRequest protocol status reports error: %d.", endRequestBody.protocolStatus);
				
				break;
			}

			if (responseHeader.type != FcgiMessage::Stdout &&
				responseHeader.type != FcgiMessage::Stderr)
				throw fastcgi_error(FcgiError::ProtocolError, "FASTCGI: only StdOut/StdErr response supported, received message type: %d.", responseHeader.type);
			
			size_t contentLen = (responseHeader.contentLengthB1 << 8) + responseHeader.contentLengthB0;
			size_t fullLength = contentLen + responseHeader.paddingLength;

						
			if (responseHeader.type == FcgiMessage::Stdout)
			{
				// check headers
				if (!headersLoaded) {
					aconnect::string dataBlock = input.read (fullLength); 
					if (dataBlock.size() != fullLength)
						throw fastcgi_error(FcgiError::ProtocolError, "FASTCGI: StdOut response corrupted.");
			
					size_t headEnd = dataBlock.find(strings::HeadersEndMark);

					if (headEnd != string::npos) {

						context.Response.Header.load (0, dataBlock.substr(0, headEnd).c_str());

						headEnd += strlen(strings::HeadersEndMark);
						context.Response.write(dataBlock.c_str() + headEnd,
							contentLen - headEnd);

						headersLoaded = true;
					}
				
				} else if (contentLen > 0) {
					size_t written  = input.readContentTo<> (contentLen, responseHeader.paddingLength, 
						&ahttp::HttpResponse::write, 
						&(context.Response));

					if (written != fullLength)
						throw fastcgi_error(FcgiError::ProtocolError, "FASTCGI: StdOut response body corrupted.");
				}

				if (contentLen == 0) // TODO: control by 'end-request-at-empty-write' param
					break;
			}
			else
			{
				aconnect::string content = input.read (fullLength); 
				if (content.size() != (size_t) fullLength)
					throw fastcgi_error(FcgiError::ProtocolError, "FASTCGI: StdErr response corrupted.");

				if (contentLen > 0) {
					std::string errDesc = aconnect::util::escapeFormatSigns (content.substr(0, contentLen));
					Globals::Log->error("FASTCGI-STDERR: %s", errDesc.c_str());

					// prepare HTML response
					errDesc = util::escapeHtml (errDesc);
					algo::replace_all (errDesc, "\n", "<br />");
					algo::replace_all (errDesc, "  ", "&nbsp;&nbsp;");
					context.processServerError(errDesc.c_str() );

				} else {
					break; // TODO: control by 'end-request-at-empty-write' param
				}
			}

		} while (responseHeader.type != fastcgi::FcgiMessage::EndRequest);


		// update Content-Type if not set
		context.setHtmlResponse();
	
	} catch (std::exception const &ex)  {
		context.Log->error ("FastCGI: exception caught (%s): %s\t Path: %s", 
			typeid(ex).name(), ex.what(),
			context.VirtualPath.c_str());
		
		context.Server->processServerError(context, 
			HttpStatus::InternalServerError, 
			ex.what());
			
	} catch (...)  {
		context.Log->error ("FastCGI: unknown exception caught, path: %s",
			context.VirtualPath.c_str());
		
		context.Server->processServerError(context);
	}
	 
	
	return true;
}

void beginRequest (ahttp::HttpContext& context, 
				   const fastcgi::SocketDescriptor& sd, 
				   const int requestId)
{
	using namespace ahttp;
	
	fastcgi::ApplicationManager::writeBeginRequestMessage (sd, requestId, false);
	
	// load server variables
	aconnect::str2str_map params = context.getCgiVariables();
	string variablesInfo = fastcgi::support::formatKeyValuePairs(params);

	// Write request params
	size_t remainBytesToWrite = variablesInfo.size();
	size_t currentPos = 0;
	fastcgi::Header header;
	
	while (remainBytesToWrite > 0) {
		int bytesToWrite = (int) aconnect::util::min2(remainBytesToWrite, fastcgi::MaxMessageContentLength);
		unsigned int padding = aconnect::util::alignInt8 (bytesToWrite) - bytesToWrite;
		
		fastcgi::support::formatFcgiHeader (header,
			fastcgi::FcgiMessage::Params,
			requestId,
			bytesToWrite,
			padding);

		fastcgi::ApplicationManager::write (sd, (fastcgi::byte_type*) &header, fastcgi::HeaderLenght);
		fastcgi::ApplicationManager::write (sd, (fastcgi::byte_type*) (variablesInfo.data() + currentPos), bytesToWrite);

		if (padding > 0)
			fastcgi::ApplicationManager::writePadding (sd, padding);
			
		currentPos += bytesToWrite;
		remainBytesToWrite -= bytesToWrite;
	}

	// Send empty FCGI_PARAMS
	fastcgi::support::formatFcgiHeader (header,
			fastcgi::FcgiMessage::Params,
			requestId);
	fastcgi::ApplicationManager::write (sd, (fastcgi::byte_type*) &header, fastcgi::HeaderLenght);

	size_t requestContentLength = context.RequestHeader.ContentLength;
	
	// Send POST content
	if (requestContentLength > 0) 
	{
		size_t padding = 0;
		const size_t postBufferSize = util::min2(fastcgi::MaxMessageContentLength, requestContentLength);
		boost::scoped_array<fastcgi::byte_type> postBuffer(new byte_type[ postBufferSize ] );
		
		while (requestContentLength > 0) {
			
			size_t partSize = context.RequestStream.read ((string_ptr) postBuffer.get(), postBufferSize );
			padding = aconnect::util::alignInt8 (partSize) - partSize;

			fastcgi::support::formatFcgiHeader (header,
				fastcgi::FcgiMessage::Stdin,
				requestId,
				partSize,
				padding);

			fastcgi::ApplicationManager::write (sd, (fastcgi::byte_type*) &header, fastcgi::HeaderLenght);
			fastcgi::ApplicationManager::write (sd, postBuffer.get(), partSize);
			if (padding > 0)
				fastcgi::ApplicationManager::writePadding (sd, padding);

			requestContentLength -= partSize;
		}
	}


	// Send empty FCGI_STDIN
	fastcgi::support::formatFcgiHeader (header,
			fastcgi::FcgiMessage::Stdin,
			requestId);
	fastcgi::ApplicationManager::write (sd, (fastcgi::byte_type*) &header, fastcgi::HeaderLenght);
	
}
