/*
This file is part of [ahttpserver] project.

Author: Artem Kustikov (kustikoff[at]tut.by)


This code is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any
damages arising from the use of this code.

Permission is granted to anyone to use this code for any
purpose, including commercial applications, and to alter it and
redistribute it freely, subject to the following restrictions:

1. The origin of this code must not be misrepresented; you must
not claim that you wrote the original code. If you use this
code in a product, an acknowledgment in the product documentation
would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and
must not be misrepresented as being the original code.

3. This notice may not be removed or altered from any source
distribution.
*/

// python_handler.cpp : Defines the entry point for the DLL application.
//

#include "aconnect/config.hpp"
#include <boost/algorithm/string.hpp>
#include <boost/thread.hpp>
#include <boost/filesystem.hpp>

#include "ahttplib.hpp"
#include "aconnect/util.hpp"
#include "aconnect/util.string.hpp"
#include "aconnect/util.file.hpp"
#include "aconnect/util.guards.hpp"

namespace algo = boost::algorithm;
namespace fs = boost::filesystem;

#include "wrappers.hpp"
#include "module.inl"

struct InterpreterRecord
{
	int	handlerIndex;
	aconnect::string uploadsDirPath;
	
	PyThreadState*		threadState;
	python::object		mainModule;
	python::dict		mainModuleDict;

	// WSGI
	aconnect::string	loaderPath;
	aconnect::string	moduleName;
	aconnect::string	loaderFilePath;

	python::object		impModule;
	bool				watchLoaderFile;
	std::time_t			loaderFileTimestamp;
	boost::mutex		reloadMutex;

	python::object		requestFunction;    // function
	
	InterpreterRecord(): 
		handlerIndex (-1),
		threadState(0),
		watchLoaderFile(false),
		loaderFileTimestamp(0)

	{
	}
};


namespace PythonGlobals
{
	// constants
	const aconnect::string Param_UploadsDir = "uploads-dir";
	const aconnect::string Param_WsgiEngine = "wsgi-engine";
	const aconnect::string Param_WatchLoaderFile = "watch-loader-file";
	

	const aconnect::string_constptr WcgiProcessingHandlerName = "process";
	const aconnect::string WcgiLoaderModuleName = "ahttpserver_wcgi_";
	
	// globals
	boost::mutex LoadMutex;
    
	static ahttp::HttpServerSettings *GlobalServerSettings = 0;

	typedef std::map<int, InterpreterRecord*> engines_map;
	engines_map RegisteredEnginesMap;

	python::dict DefaultWsgiEnviron;
}


HANDLER_EXPORT bool initPlugin  (const aconnect::str2str_map& params, int handlerIndex,
								  ahttp::HttpServer* server);
HANDLER_EXPORT void destroyPlugin ();
HANDLER_EXPORT bool processHandlerRequest (ahttp::HttpContext& context, int handlerIndex);

void executeScriptDirectly (aconnect::string_constref scriptPath, ahttp::HttpContext& context, InterpreterRecord *interpreterRecord, PyThreadState *ts);

void processWcgiCall (ahttp::HttpContext& context, InterpreterRecord *interpreterRecord, PyThreadState *ts);

void releaseExecutionThread (PyThreadState *ts);
bool initPythonGlobals(ahttp::HttpServerSettings* serverSettings);
bool loadRequestsHandler (aconnect::string_constptr path, InterpreterRecord* record);

//////////////////////////////////////////////////////////////////////////
//	
//	Windows related stuff
#if defined (WIN32)
BOOL APIENTRY DllMain( HMODULE hModule,
					  DWORD  ul_reason_for_call,
					  LPVOID lpReserved
					  )
{
	if (ul_reason_for_call == DLL_PROCESS_ATTACH)
	{
		// Don't need to be called for new threads
		DisableThreadLibraryCalls ((HMODULE) hModule);
	}
	return TRUE;
}
#endif



/* 
*	Handler initialization function - return true if initialization performed sucessfully
*/
HANDLER_EXPORT bool initPlugin  (const aconnect::str2str_map& params, 
								  int handlerIndex,
								  ahttp::HttpServer* server)
{
	assert (server);
	assert (server->GlobalSettings());
	assert (server->Log());
	assert (server->GlobalSettings()->logger());
	
	boost::mutex::scoped_lock lock(PythonGlobals::LoadMutex);
    
	// first record - prepare globals
    if (PythonGlobals::GlobalServerSettings == 0) 
	{
		PythonGlobals::GlobalServerSettings = server->GlobalSettings();
		
		if ( !initPythonGlobals (PythonGlobals::GlobalServerSettings) )
			return false;
	}

	assert (PythonGlobals::RegisteredEnginesMap.find(handlerIndex) == PythonGlobals::RegisteredEnginesMap.end());

	std::auto_ptr<InterpreterRecord> record(new InterpreterRecord());

	record->handlerIndex = handlerIndex;

	// load registration settings
	aconnect::str2str_map::const_iterator paramsIt = params.find(PythonGlobals::Param_UploadsDir);
	
    if (paramsIt != params.end()) {
		record->uploadsDirPath = paramsIt->second;
		if (!fs::exists (record->uploadsDirPath)) 
		{
			try {
				fs::create_directories(record->uploadsDirPath);
			} catch (...) {
				server->Log()->error ("Python: uploads directory creation failed", PythonGlobals::Param_UploadsDir.c_str() );
				return false;
			}
		} 
	} 

	// Prepare new interpreter
	PyEval_AcquireLock();

	record->threadState = Py_NewInterpreter();
	if (record->threadState == 0) {
		server->Log()->error ("Python: Py_NewInterpreterfailed, out of memory");
		return false;
	}

	// process in new interpreter
	PyThreadState* savedState = PyThreadState_Swap (record->threadState);
	
	// Register the module with the interpreter
	if (PyImport_AppendInittab( (char*) "python_handler", initpython_handler) == -1)
	{
		PythonGlobals::GlobalServerSettings->logger()->error ("Failed to register 'python_handler' in the interpreter's built-in modules");
		return false;
	}
			
	// Retrieve the main module's namespace
	record->mainModule = python::import("__main__");
	record->mainModuleDict = python::dict (record->mainModule.attr("__dict__"));
	
	// register classes
	python::import ("python_handler");

	// load requests processor
	paramsIt = params.find(PythonGlobals::Param_WsgiEngine);
	if (paramsIt != params.end()) {
		record->loaderPath = paramsIt->second;
		PythonGlobals::GlobalServerSettings->updateAppLocationInPath(record->loaderPath);

		if ( !loadRequestsHandler(record->loaderPath.c_str(), 
				record.get()) )
			return false;

		if (!record->loaderFilePath.empty()) 
			record->watchLoaderFile = aconnect::util::getItemFromMapBool(params, PythonGlobals::Param_WatchLoaderFile, record->watchLoaderFile);

		// WSGI application exist - prepare environment
		if ( python::len (PythonGlobals::DefaultWsgiEnviron.keys()) == 0) 
		{
			PythonGlobals::DefaultWsgiEnviron["wsgi.version"] = python::make_tuple(1,0);
			PythonGlobals::DefaultWsgiEnviron["wsgi.multithread"] = true;
			PythonGlobals::DefaultWsgiEnviron["wsgi.multiprocess"] = false;
			PythonGlobals::DefaultWsgiEnviron["wsgi.run_once"] = false;
		}

	}
	
	PythonGlobals::RegisteredEnginesMap[handlerIndex] = record.release();
	
    PyThreadState_Swap(savedState);
	PyEval_ReleaseLock ();

	return true;
}


HANDLER_EXPORT void destroyPlugin  ()
{
	boost::mutex::scoped_lock lock(PythonGlobals::LoadMutex);

	if ( Py_IsInitialized() )
	{
		PythonGlobals::engines_map::iterator it = PythonGlobals::RegisteredEnginesMap.begin();
		
		while (it != PythonGlobals::RegisteredEnginesMap.end()) 
		{
			PyEval_AcquireLock();
			
			PyThreadState_Swap(it->second->threadState );
								
			PyThreadState_Swap (0);
			PyEval_ReleaseLock();

			PyThreadState_Clear(it->second->threadState);
			PyThreadState_Delete(it->second->threadState); 
			Py_XDECREF (it->second->threadState);

			delete it->second;

			++it;
		}

		PyInterpreterState* is = PyInterpreterState_New();

		if (is) 
		{
			PyThreadState* ts = PyThreadState_New(is);
			PyThreadState_Swap(ts);
			Py_Finalize();
			PyThreadState_Swap(NULL);
		}
	}
	
	PythonGlobals::GlobalServerSettings = 0;
}

/* 
*	Main request processing function - return false if request should
*	be processed by other handlers or by server (true: request completed)
*/
HANDLER_EXPORT bool processHandlerRequest (ahttp::HttpContext& context, 
										   int handlerIndex)
{
	using namespace ahttp;
	using namespace aconnect;
	
	PythonGlobals::engines_map::iterator iter = PythonGlobals::RegisteredEnginesMap.find(handlerIndex);
	if (iter == PythonGlobals::RegisteredEnginesMap.end()) {
		context.processServerError("Python handler registarion loading failed");
		return true;
	}

	InterpreterRecord *interpreterRecord = iter->second;

	// init context
	if (!interpreterRecord->uploadsDirPath.empty())
		context.UploadsDirPath = interpreterRecord->uploadsDirPath;
	
	if (!context.isClientConnected())
		return true;
	
	bool isDirectCall = (interpreterRecord->requestFunction.ptr() == Py_None);

	// check direct access
	if (isDirectCall) {
		if ( !util::fileExists (context.FileSystemPath.string()) ) {
			context.Server->processError404 (context);
			return true;
		}
	}
	
	PyThreadState *ts = 0;
	util::scoped_param_guard < void (*) (PyThreadState *), PyThreadState*> guardTs (releaseExecutionThread, ts);
	
	try 
	{
		// get the global lock
		PyEval_AcquireLock();
		// get a reference to the PyInterpreterState
		PyInterpreterState * interpreterState = interpreterRecord->threadState->interp;

		// create a thread state object for this thread
		ts = PyThreadState_New (interpreterState);

		if ( !ts ) {
            PyEval_ReleaseLock();
			throw std::runtime_error("Python: thread state object creation failed - out of memory.");
        }

		PyThreadState_Swap(ts);

		if (isDirectCall) {
			executeScriptDirectly (context.FileSystemPath.file_string(), context, interpreterRecord, ts);
		
		} else {
			
			if (interpreterRecord->watchLoaderFile) 
			{
				boost::mutex::scoped_lock watchLock(interpreterRecord->reloadMutex);
				std::time_t timeStamp = fs::last_write_time (interpreterRecord->loaderFilePath);

				if (timeStamp != interpreterRecord->loaderFileTimestamp) 
				{
					python::object loaderModule = interpreterRecord->impModule.attr("load_source")(interpreterRecord->moduleName, interpreterRecord->loaderFilePath);
					
					interpreterRecord->requestFunction = loaderModule.attr(PythonGlobals::WcgiProcessingHandlerName);
					interpreterRecord->loaderFileTimestamp = timeStamp;
				}
			}

            // TODO: update sys.argv (see mod_wsgi)

			processWcgiCall (context, interpreterRecord, ts);
		}

        guardTs.release();

		context.setHtmlResponse();

        return true;

	} catch (python::error_already_set const &)  {
		
		try
		{
			string errDesc = loadPythonError ();
			context.Log->errorSafe(errDesc.c_str());
			
			// prepare HTML response
			errDesc = util::escapeHtml (errDesc);
			algo::replace_all (errDesc, "\n", "<br />");
			algo::replace_all (errDesc, "  ", "&nbsp;&nbsp;");
			context.Server->processServerError(context, 500, aconnect::util::escapeFormatSigns(errDesc).c_str() );
		}
		catch (...)
		{
			context.Log->error("Python: loading Python error failed, file: %s", 
				context.FileSystemPath.string().c_str());

			context.Server->processServerError(context, ahttp::HttpStatus::InternalServerError, 
				"Python: loading Python error failed, file: %s", 
				context.FileSystemPath.string().c_str());
		}

	} catch (std::exception const &ex)  {
		context.Log->error ("Python: exception caught (%s): %s, file: %s", 
			typeid(ex).name(), ex.what(),
			context.FileSystemPath.string().c_str());
		
		context.Server->processServerError(context);
			
	} catch (...)  {
		context.Log->error ("Python: unknown exception caught, file: %s",
			context.FileSystemPath.string().c_str());
		
		context.Server->processServerError(context);
	}
	
    // exception occured

	return true;
}

//==================================================================//
//	Internal module logic											//
//==================================================================//


void releaseExecutionThread (PyThreadState *ts) 
{
	if (ts)
	{
        // PyThreadState_Clear (ts);
        // PyEval_ReleaseThread (ts);
        // PyThreadState_Delete (ts);
        // PyEval_ReleaseLock();

		// clear the thread state
		ts = PyThreadState_Swap(0);

		PyEval_ReleaseLock();
		
		if (ts) 
		{
			PyThreadState_Clear(ts);
			PyThreadState_Delete(ts);
		}
	}
}

bool initPythonGlobals(ahttp::HttpServerSettings* serverSettings)
{
	assert (serverSettings);

	Py_OptimizeFlag = 0;
		
	// INVESTIGATE: set home and other things - check mod_wsgi
	// Py_SetPythonHome((char *)config->python_home);

	// Init Python interpreter in multithreaded mode
	Py_Initialize ();

	if ( 0 == Py_IsInitialized()) {
		serverSettings->logger()->error ("Python interpreter was not initialized correctly");
		return false;
	}

	PyEval_InitThreads();

	PyObject *serverPath = PyString_FromString (serverSettings->AppLocation.c_str());
	PySys_SetObject( (char*) "executable", serverPath);
	Py_DecRef(serverPath);

	PyThreadState_Swap(0);
	PyEval_ReleaseLock();

	return true;
}

bool loadRequestsHandler (aconnect::string_constptr loaderPath, InterpreterRecord* record)
{

	if (record->requestFunction.ptr() != Py_None)
		record->requestFunction = python::object();

	try 
	{
		python::object loaderModule;
		// file-loader - import by path
		if (algo::ends_with(loaderPath, ".py", algo::is_iequal()))
		{
			fs::path filePath (loaderPath, fs::native);
			
			if (!fs::exists(filePath)) {

				filePath = fs::path(PythonGlobals::GlobalServerSettings->getRootDirSettings().realPath, fs::native) / filePath;

				if (!fs::exists(filePath))
					filePath = fs::path(PythonGlobals::GlobalServerSettings->AppLocation, fs::native) / fs::path(loaderPath, fs::native);
				
				if (!fs::exists(filePath)) {
					PythonGlobals::GlobalServerSettings->logger()->error ("Python requests loader script not found, %s",
						loaderPath);
					return false;
				}
			}

			record->impModule = python::import ("imp");
			record->moduleName = PythonGlobals::WcgiLoaderModuleName + boost::lexical_cast<std::string>(record->handlerIndex);
			record->loaderFilePath = filePath.file_string();
			record->loaderFileTimestamp = fs::last_write_time (filePath);

			loaderModule = record->impModule.attr("load_source")(record->moduleName, record->loaderFilePath);
		
		// Python module from PYTHONPATH
		} else {
			loaderModule = python::import (loaderPath);
			record->moduleName = loaderPath;
		}

		record->requestFunction = loaderModule.attr(PythonGlobals::WcgiProcessingHandlerName);
	}
	catch (python::error_already_set) 
	{
		aconnect::string errDesc = loadPythonError ();
		PythonGlobals::GlobalServerSettings->logger()->errorSafe(errDesc.c_str());
		return false;
	}
	catch (std::exception &ex) 
	{
		PythonGlobals::GlobalServerSettings->logger()->error(ex);
		return false;
	}

	return true;
}


/**
 * Execute python code directly from .py file
 */
void executeScriptDirectly (aconnect::string_constref scriptPath, ahttp::HttpContext& context, InterpreterRecord *interpreterRecord,  PyThreadState *ts )
{
	using namespace python;
	
	dict global (interpreterRecord->mainModuleDict.copy ());
	dict local (interpreterRecord->mainModuleDict.copy ());
	
	// prepare globals
	python::manage_new_object::apply<HttpContextWrapper*>::type converter;
	handle<> wrapperHandle ( converter( new HttpContextWrapper(&context) ) );
	
	local["http_context"] = wrapperHandle;
	
	PyObject *pyfile = PyFile_FromString (const_cast<char*>(scriptPath.c_str()), (char*) "r");
	python::handle<> file (pyfile);
	FILE* fptr = PyFile_AsFile(file.get());
	
	PyObject* result = PyRun_File (fptr,
		scriptPath.c_str(),
		Py_file_input,
		global.ptr(), 
		local.ptr());
	
	if (!result) 
		throw_error_already_set();
}

void processWcgiCall (ahttp::HttpContext& context, InterpreterRecord *interpreterRecord, PyThreadState *ts)
{
	using namespace python;
	using namespace ahttp::strings::ServerVariables;
	
	manage_new_object::apply<HttpContextWrapper*>::type contextConverter;
	manage_new_object::apply<RequestWrapper*>::type requestConverter;

	handle<> contextHandle ( contextConverter( new HttpContextWrapper (&context) ) );
	handle<> requestHandle ( requestConverter( new RequestWrapper (&context) ) );
	
	dict environ (PythonGlobals::DefaultWsgiEnviron.copy());
	
	// write server variables
    {
		aconnect::str2str_map params = context.getCgiVariables();
		
		// maybe remove SCRIPT_NAME/SCRIPT_FILENAME?
		// write PATH_INFO required by WSGI

		assert(params[SCRIPT_NAME].size());
		assert(params[APPL_PATH].size());
		params[PATH_INFO] = params[SCRIPT_NAME].substr(params[APPL_PATH].size() - 1); 

		aconnect::str2str_map::const_iterator varIter = params.begin();
		while (varIter != params.end()) {
			environ[varIter->first] = varIter->second;
			++varIter;
		}
	}

	// register WSGI callbacks
	environ["wsgi.input"] = requestHandle;
	environ["wsgi.errors"] = python::handle<>(Py_None);
	environ["wsgi.url_scheme"] = "http";
	
	PyObject * callRes = PyEval_CallFunction (interpreterRecord->requestFunction.ptr(),
		"OO",
		environ.ptr(),
		contextHandle.get());
	
	// object response = interpreterRecord->requestFunction(environ, contextHandle);  //__call__ => start_response
	
	assert (callRes != Py_None);
	
	object response = object(python::handle<>(callRes));
	object iter = response.attr("__iter__")();

	try 
	{
		while (true) {
			object dataPtr = iter.attr("next")();
			std::string data = extract<std::string>(dataPtr);

			PyThreadStateGuard stateGuard;
			context.Response.writeString (data);
		}
	} catch (python::error_already_set const &)  {
		if (0 == PyErr_ExceptionMatches(PyExc_StopIteration))
			throw;
	}
	
	if ( !context.RequestStream.isRead())
		context.loadPostParams();

}

// #include "d:/lib/boost_1_38_0/libs/python/test/module_tail.cpp" 
// #include "/media/work/lib/boost_1_38_0/libs/python/test/module_tail.cpp" 
