
// each test module could contain no more then one 'main' file with init function defined
// alternatively you could define init function yourself

#include <boost/test/unit_test.hpp>
#include <boost/filesystem.hpp>

#include <fstream>

#include "aconnect/util.hpp"
#include "aconnect/logger.hpp"
#include "aconnect/client.hpp"

namespace fs = boost::filesystem;

namespace Globals
{
	extern aconnect::string ApplicationDirectory;
	aconnect::string_constant LogDirectory = "log_test";
}
//==================================================================//

struct F {
    F() { 

		testDir = fs::complete( fs::path(Globals::LogDirectory), fs::path(Globals::ApplicationDirectory));

		BOOST_TEST_MESSAGE( "setup fixture" );
		clearLoggerDirectory ();
	}
    
	~F() { 
		BOOST_TEST_MESSAGE( "teardown fixture" ); 
	}

	fs::path testDir;

private:
	void clearLoggerDirectory ()
	{
		if (!fs::exists(testDir)) {
			fs::create_directories (testDir);
			return;
		}
		
		fs::directory_iterator endTter;
		for ( fs::directory_iterator dirIter (testDir);
			dirIter != endTter;
			++dirIter )
		{
				fs::remove(dirIter->path());
		}
	}
   
};

//==================================================================//

// struct F is going to be used as a fixture for all test cases in this test suite
BOOST_FIXTURE_TEST_SUITE (IO_tests, F )
 
BOOST_AUTO_TEST_CASE( Background_Logger )
{
	using namespace aconnect;
	BackgroundFileLogger logger;
	
	fs::path logFilePath = fs::complete( fs::path("bg{timestamp}.log"), testDir);
	logger.init (Log::Debug, logFilePath.file_string().c_str());

	string logFile = logger.currentFilePath();

	BOOST_CHECK ( fs::exists(logFile) );

	logger.debug ("Debug");
	logger.info ("Info");
	logger.warn ("Warn");
	logger.error ("Error");
	logger.critical ("Critical");

	logger.destroy();
	
	std::fstream input(logFile.c_str(), std::ios::in | std::ios::binary);	
	
	const int lineSize = 255;
	aconnect::char_type line[lineSize];
	
	input.getline (line, 255);
	BOOST_CHECK ( NULL != strstr(line, "Debug") );

	input.getline (line, 255);
	BOOST_CHECK ( NULL != strstr(line, "Info") );

	input.getline (line, 255);
	BOOST_CHECK ( NULL != strstr(line, "Warn") );

	input.getline (line, 255);
	BOOST_CHECK ( NULL != strstr(line, "Error") );

	input.getline (line, 255);
	BOOST_CHECK ( NULL != strstr(line, "Critical") );
}


BOOST_AUTO_TEST_CASE( Background_Logger_Volume_Test )
{
	using namespace aconnect;
	BackgroundFileLogger logger;

	const int linesCount = 1024 * 100;
	
	fs::path logFilePath = fs::complete( fs::path("bg_mass{timestamp}.log"), testDir);
	logger.init (Log::Warning, logFilePath.file_string().c_str(), linesCount * 64 /* line size */);

	string logFile = logger.currentFilePath();

	BOOST_CHECK ( fs::exists(logFile) );

	

	for (int ndx=0; ndx < linesCount; ++ndx) {
		logger.debug ("Debug");
		logger.info ("Info");
		logger.warn ("Line %d", ndx);
	}

	logger.destroy();
	
	std::fstream input(logFile.c_str(), std::ios::in | std::ios::binary);	
	
	const int lineSize = 255;
	const int markSize = 32;
	aconnect::char_type line[lineSize];
	aconnect::char_type mark[markSize];

	
	for (int ndx=0; ndx<linesCount; ++ndx) {

		int formatted = snprintf (mark, markSize, "Line %d", ndx);
		BOOST_CHECK ( formatted > 0 );
		 
		input.getline (line, 255);
		BOOST_REQUIRE ( NULL != strstr(line, mark) );
	}
}


BOOST_AUTO_TEST_SUITE_END()
